"""Tools to help handling QUA loops.

Content:
    - from_array: Function parametrizing the QUA `for_` loop from a python array.
    - qua_arange: Function parametrizing the QUA `for_` loop from the numpy.arange() syntax.
    - qua_linspace: Function parametrizing the QUA `for_` loop from the numpy.linspace() syntax.
    - qua_logspace: Function parametrizing the QUA `for_` loop from the numpy.logspace() syntax.
"""

import numpy as np
import warnings
from qm.qua import Cast
from qm.qua import Variable


def from_array(var, array):
    """Function parametrizing the QUA `for_` loop from a python array.

    :param var: the QUA variable that will be looped over (int or fixed).
    :param array: a Python list or numpy array containing the values over which `a` will be looping. The spacing must be even in linear or logarithmic scales and it cannot be a QUA array.
    :return: QUA for_ loop parameters (var, init, cond, update) as defined in https://qm-docs.qualang.io/api_references/qua/dsl_main?highlight=for_#qm.qua.for_.
    """

    # Check for array length
    if len(array) == 0:
        raise Exception("The array must be of length > 0.")
    elif len(array) == 1:
        return var, array[0], var <= array[0], var + 1
    # Check QUA vs python variables
    if not isinstance(var, Variable):
        raise Exception("The first argument must be a QUA variable.")
    if (not isinstance(array[0], (np.generic, int, float))) or (isinstance(array[0], bool)):
        raise Exception("The array must be an array of python variables.")
    # Check array increment
    if np.isclose(np.std(np.diff(array)), 0):
        increment = "lin"
    elif np.isclose(np.std(array[1:] / array[:-1]), 0, atol=1e-3):
        increment = "log"
    else:
        raise Exception(
            "The spacing of the input array must be even in linear or logarithmic scales. Please use `for_each_()` for arbitrary scans."
        )

    # Get the type of the specified QUA variable
    start = array[0]
    stop = array[-1]
    # Linear increment
    if increment == "lin":
        step = array[1] - array[0]

        if var.is_int():
            # Check that the array is an array of int with integer increments
            if not float(step).is_integer() or not float(start).is_integer() or not float(stop).is_integer():
                raise Exception("When looping over a QUA int variable, the step and array elements must be integers.")
            # Generate the loop parameters for positive and negative steps
            if step > 0:
                return var, int(start), var <= int(stop), var + int(step)
            else:
                return var, int(start), var >= int(stop), var + int(step)

        elif var.is_fixed():
            # Check for fixed number overflows
            if not (-8 <= start < 8) and not (-8 <= stop < 8):
                raise Exception("fixed numbers are bounded to [-8, 8).")

            # Generate the loop parameters for positive and negative steps
            if step > 0:
                return (
                    var,
                    float(start),
                    var < float(stop) + float(step) / 2,
                    var + float(step),
                )
            else:
                return (
                    var,
                    float(start),
                    var > float(stop) + float(step) / 2,
                    var + float(step),
                )
        else:
            raise Exception(
                "This variable type is not supported. Please use a QUA 'int' or 'fixed' or contact a QM member for assistance."
            )
    # Logarithmic increment
    elif increment == "log":
        step = array[1] / array[0]

        if var.is_int():
            warnings.warn(
                "When using logarithmic increments with QUA integers, the resulting values will slightly differ from the ones in numpy.logspace() because of rounding errors. \n Please use the get_equivalent_log_array() function to get the exact values taken by the QUA variable and note that the number of points may also change."
            )
            if step > 1:
                if int(round(start) * float(step)) == int(round(start)):
                    raise ValueError(
                        "Two successive values in the scan are equal after being cast to integers which will make the QUA for_ loop fail. \nEither increase the logarithmic step or use for_each_(): https://docs.quantum-machines.co/1.1.6/qm-qua-sdk/docs/Guides/features/?h=for_ea#for_each."
                    )
                else:
                    return (
                        var,
                        round(start),
                        var < round(stop) * np.sqrt(float(step)),
                        Cast.mul_int_by_fixed(var, float(step)),
                    )
            else:
                return (
                    var,
                    round(start),
                    var > round(stop) / np.sqrt(float(step)),
                    Cast.mul_int_by_fixed(var, float(step)),
                )

        elif var.is_fixed():
            # Check for fixed number overflows
            if not (-8 <= start < 8) and not (-8 <= stop < 8):
                raise Exception("fixed numbers are bounded to [-8, 8).")

            if step > 1:
                return (
                    var,
                    float(start),
                    var < float(stop) * np.sqrt(float(step)),
                    var * float(step),
                )
            else:
                return (
                    var,
                    float(start),
                    var > float(stop) * np.sqrt(float(step)),
                    var * float(step),
                )
        else:
            raise Exception(
                "This variable type is not supported. Please use a QUA 'int' or 'fixed' or contact a QM member for assistance."
            )


def qua_arange(var, start, stop, step):
    """Function parametrizing the QUA `for_` loop from the numpy.arange() syntax.

    :param var: the QUA variable that will be looped over (int or fixed).
    :param start: Start of interval. The interval includes this value. Must be a Python variable.
    :param stop: End of interval. The interval does not include this value. Must be a Python variable.
    :param step: Spacing between values. Must be a Python variable.
    :return: QUA for_ loop parameters (var, init, cond, update) as defined in https://qm-docs.qualang.io/api_references/qua/dsl_main?highlight=for_#qm.qua.for_.
    """
    # Check QUA vs python variables
    if not isinstance(var, Variable):
        raise Exception("The first argument must be a QUA variable.")
    if (
        (not isinstance(start, (np.generic, int, float)))
        or (not isinstance(stop, (np.generic, int, float)))
        or (not isinstance(step, (np.generic, int, float)))
        or (isinstance(start, bool))
        or (isinstance(stop, bool))
        or (isinstance(step, bool))
    ):
        raise Exception("The for loop arguments must be python variables.")
    if float(step).is_integer():
        stop_condition = stop
    else:
        # Check for fixed number overflows
        if not (-8 <= start < 8) and not (-8 <= stop < 8):
            raise Exception("fixed numbers are bounded to [-8, 8).")
        if ((stop - start) / step).is_integer():
            stop_condition = stop - step / 2
        else:
            stop_condition = stop

        if np.abs((stop - start) / step * 2**-28) >= np.abs(step):
            raise Exception("The required accuracy is to large:N * (2 ** -28) > step/2, please contact a QM member.")

    if step > 0:
        return var, start, var < stop_condition, var + step
    else:
        return var, start, var > stop_condition, var + step


def qua_linspace(var, start, stop, num):
    """Function parametrizing the QUA `for_` loop from the numpy.linspace() syntax.

    :param var: the QUA variable that will be looped over (int or fixed).
    :param start: The starting value of the sequence. The interval includes this value. Must be a Python variable.
    :param stop: The end value of the sequence. The interval includes this value within the 2**-28 fixed point accuracy. Must be a Python variable.
    :param num: Number of samples to generate. Must be a Python variable.
    :return: QUA for_ loop parameters (var, init, cond, update) as defined in https://qm-docs.qualang.io/api_references/qua/dsl_main?highlight=for_#qm.qua.for_.
    """
    # Check QUA vs python variables
    if not isinstance(var, Variable):
        raise Exception("The first argument must be a QUA variable.")
    if (
        (not isinstance(start, (np.generic, int, float)))
        or (not isinstance(stop, (np.generic, int, float)))
        or (not isinstance(num, (np.generic, int, float)))
        or (isinstance(start, bool))
        or (isinstance(stop, bool))
        or (isinstance(num, bool))
    ):
        raise Exception("The for loop arguments must be python variables.")
    # Check for fixed number overflows
    if not (-8 <= start < 8) and not (-8 <= stop < 8):
        raise Exception("fixed numbers are bounded to [-8, 8).")
    # Check if num is integer
    if not float(num).is_integer():
        raise Exception("The number of samples must be a python integer.")

    if num > 1:
        step = (stop - start) / (num - 1)
    elif num == 1:
        step = (stop - start) * 2
    else:
        step = stop - start

    if step > 0:
        return var, start, var < stop + step / 2, var + step
    else:
        return var, start, var > stop + step / 2, var + step


def qua_logspace(var, start, stop, num):
    """Function parametrizing the QUA `for_` loop from the numpy.logspace() syntax.

    :param var: the QUA variable that will be looped over (fixed).
    :param start: The starting value of the sequence as 10**start in linear space. The interval includes this value. Must be a Python variable.
    :param stop: The end value of the sequence as 10**stop in linear space. The interval includes this value within the 2**-28 fixed point accuracy. Must be a Python variable.
    :param num: Number of samples to generate. Must be a Python variable.
    :return: QUA for_ loop parameters (var, init, cond, update) as defined in https://qm-docs.qualang.io/api_references/qua/dsl_main?highlight=for_#qm.qua.for_.
    """
    # Check QUA vs python variables
    if not isinstance(var, Variable):
        raise Exception("The first argument must be a QUA variable.")
    if (
        (not isinstance(start, (np.generic, int, float)))
        or (not isinstance(stop, (np.generic, int, float)))
        or (not isinstance(num, (np.generic, int, float)))
        or (isinstance(start, bool))
        or (isinstance(stop, bool))
        or (isinstance(num, bool))
    ):
        raise Exception("The for loop arguments must be python variables.")
    # Check if num is integer
    if not float(num).is_integer():
        raise Exception("The number of samples must be a python integer.")

    if num > 1:
        step = (10**stop / 10**start) ** (1 / (num - 1))
    elif num == 1:
        step = (10**stop / 10**start) ** 1
    else:
        raise Exception("`num` must be greater than 0.")

    # Get the type of the specified QUA variable
    if var.is_int():
        warnings.warn(
            "When using logarithmic increments with QUA integers, the resulting values will slightly differ from the ones in numpy.logspace() because of rounding errors. \n Please use the get_equivalent_log_array() function to get the exact values taken by the QUA variable and note that the number of points may also change."
        )
        if step > 1:
            if int(round(10**start) * float(step)) == int(round(10**start)):
                raise ValueError(
                    "Two successive values in the scan are equal after being cast to integers which will make the QUA for_ loop fail. \nEither increase the logarithmic step or use for_each_(): https://docs.quantum-machines.co/1.1.6/qm-qua-sdk/docs/Guides/features/?h=for_ea#for_each."
                )
            else:
                return (
                    var,
                    round(10**start),
                    var < round(10**stop * np.sqrt(step)),
                    Cast.mul_int_by_fixed(var, float(step)),
                )
        else:
            return (
                var,
                round(10**start),
                var > round(10**stop / np.sqrt(step)),
                Cast.mul_int_by_fixed(var, float(step)),
            )
    elif var.is_fixed():
        # Check for fixed number overflows
        if not (-8 <= 10**start < 8) and not (-8 <= 10**stop < 8):
            raise Exception("fixed numbers are bounded to [-8, 8).")

        if step > 1:
            return (
                var,
                10**start,
                (var < 10**stop * np.sqrt(step)) & (10**start <= var),
                var * step,
            )
        else:
            return (
                var,
                10**start,
                (var > 10**stop * np.sqrt(step)) & (10**start >= var),
                var * step,
            )


def get_equivalent_log_array(log_array):
    """Function returning the values taken by the QUA int variable within the logarithmic QUA `for_` loop.
    Because of rounding errors occuring with QUA integers, these values are not exactly the ones given by `numpy.logspace()`.
    Note that the number of points may also change.

    :param log_array: a Python list or numpy array containing the values parametrizing the QUA `for_` loop. The spacing must be even in logarithmic scale and it cannot be a QUA array.
    :return: numpy array containing the values taken by the QUA int variable within the logarithmic QUA `for_` loop.
    """
    a_log = []
    aprev = round(log_array[0])
    step = np.mean(np.array(log_array[1:]) / np.array(log_array[:-1]))

    if step > 1:
        end = log_array[-1] * np.sqrt(step)
        while aprev < end:
            a_log.append(aprev)
            if aprev == int(aprev * step):
                raise ValueError(
                    "Two successive values in the scan are equal after being cast to integers which will make the QUA for_ loop fail. \nEither increase the logarithmic step or use for_each_() instead of from_array(): https://docs.quantum-machines.co/1.1.6/qm-qua-sdk/docs/Guides/features/?h=for_ea#for_each."
                )
            else:
                aprev = int(aprev * step)
    else:
        end = log_array[-1] / np.sqrt(step)
        while aprev > end:
            a_log.append(aprev)
            if aprev == int(aprev * step):
                raise ValueError(
                    "Two successive values in the scan are equal after being cast to integers which will make the QUA for_ loop fail. \nEither increase the logarithmic step or use for_each_() instead of from_array(): https://docs.quantum-machines.co/1.1.6/qm-qua-sdk/docs/Guides/features/?h=for_ea#for_each."
                )
            else:
                aprev = int(aprev * step)
    return np.array(a_log)



import logging

from qm import SimulatorControllerSamples

from .helpers import copy_nested_dict, iterate_nested_dict, update_nested_dict
from .data_processor import DataProcessor


logger = logging.getLogger(__name__)


class SimulatorControllerSamplesSaver(DataProcessor):
    def process(self, data):
        processed_data = copy_nested_dict(data)
        for keys, val in iterate_nested_dict(d=data):
            if not isinstance(val, SimulatorControllerSamples):
                continue

            try:
                serialised_samples = {
                    # analog structure: {"{int}-{int}: array}
                    "analog": dict(val.analog),
                    "digital": dict(val.digital),
                    "analog_sampling_rate": dict(getattr(val, "analog_sampling_rate", {})),
                }
                update_nested_dict(processed_data, keys, serialised_samples)
            except Exception:
                logger.warning(f"Could not serialise simulator controller samples for {keys}")
                continue

        return processed_data



from typing import Union, List, Tuple

ElementType = Union[int, str]
QubitType = Union[int, str]
QubitPairType = Union[Tuple[int, int], Tuple[str, str], Tuple[int, str]]

ElementsType = Union[ElementType, List[ElementType]]
QubitsType = Union[QubitType, List[QubitType]]
QubitPairsType = Union[QubitPairType, List[QubitPairType]]



from pathlib import Path
import json
from typing import Dict, Any
import logging

from qm.waveform_report import WaveformReport

from .helpers import copy_nested_dict, iterate_nested_dict, update_nested_dict
from .data_processor import DataProcessor

logger = logging.getLogger(__name__)


class WaveformReportSaver(DataProcessor):
    nested_separator: str = "."

    def __init__(self):
        self.wf_reports: Dict[Path, WaveformReport] = {}
        self.samples: Dict[Path, Any] = {}

    def process(self, data):
        processed_data = copy_nested_dict(data)
        for keys, val in iterate_nested_dict(data):
            if not isinstance(val, WaveformReport):
                continue

            path = Path(self.nested_separator.join(keys) + ".json")

            wf_report = val

            self.wf_reports[path] = wf_report
            update_nested_dict(processed_data, keys, f"./{path}")

            try:
                parent_data = processed_data
                for key in keys[:-1]:
                    parent_data = parent_data[key]
                if "samples" in parent_data:
                    self.samples[path] = parent_data["samples"]
                else:
                    logger.warning(f"Waveform report in {path} did not have 'samples' key, resorting to default.")
                    pass
            except Exception:
                logger.warning(f"Could not extract waveform report samples from {path}")
                pass

        return processed_data

    def post_process(self, data_folder: Path):
        for path, wf_report in self.wf_reports.items():
            wf_json = json.dumps(wf_report.to_dict(), indent=4)
            wf_path = data_folder / path
            wf_path.write_text(wf_json)

            wf_report.job_id

            samples = self.samples[path] if path in self.samples else None
            wf_report.create_plot(samples=samples, plot=False, save_path=str(wf_path))



import numpy as np
from qm.octave.calibration_db import CalibrationDB
from typing import Union
from qm import QmJob
from qm.jobs.running_qm_job import RunningQmJob
from qm import QuantumMachine
from qm.octave.octave_mixer_calibration import AutoCalibrationParams


def get_calibration_parameters_from_db(
    path_to_database: str,
    config: dict,
    element: str,
    LO: float,
    IF: float,
    gain: float,
    verbose_level: int = 2,
) -> dict:
    """Returns the most up-to-date correction parameters in the database, located at the specified path_to_database, for the
    specified values of the Octave LO frequency, intermediate frequency and Octave gain.

    The correction parameters are returned in a dictionary of the form:
    ```{"offsets": {"I": I_offset, "Q": Q_offset}, "correction_matrix": correction}```.

    If the correction parameters are not found in the database:
      * verbose_level=2 will raise a warning which will block the execution.
      * verbose_level=1 will print a warning which will not block the execution  and the returned dictionary will be empty.
      * verbose_level=0 will not print anything and the returned dictionary will be empty.

    :param path_to_database: direct path to the calibration database (without '/calibration_db.json').
    :param config: the OPX config dictionary.
    :param element: the element to get the correction parameters for.
    :param LO: the LO frequency in Hz.
    :param IF: the intermediate frequency in Hz.
    :param gain: the Octave gain.
    :param verbose_level: set the type of messages printed in the Python console. Default is 2.
    :return: dictionary containing the 'I' and 'Q' offsets and the correction_matrix.
    """
    oct_name = config["elements"][element]["RF_inputs"]["port"][0]
    channel = config["elements"][element]["RF_inputs"]["port"][1]
    cal_db = CalibrationDB(path_to_database)
    lo_cal = cal_db.get_lo_cal((oct_name, channel), LO, gain)
    if_cal = cal_db.get_if_cal((oct_name, channel), LO, gain, IF)

    param = {"offsets": {}, "correction_matrix": ()}
    if lo_cal is not None:
        if hasattr(lo_cal, "get_i0"):
            param["offsets"]["I"] = lo_cal.get_i0()
        else:
            param["offsets"]["I"] = lo_cal.i0
        if hasattr(lo_cal, "get_q0"):
            param["offsets"]["Q"] = lo_cal.get_q0()
        else:
            param["offsets"]["Q"] = lo_cal.q0

        if if_cal is not None:
            if hasattr(if_cal, "get_correction"):
                param["correction_matrix"] = if_cal.get_correction()
            else:
                param["correction_matrix"] = if_cal.correction

        else:
            if verbose_level == 2:
                raise Warning(
                    f"No correction parameters were found for the provided set of LO frequency, IF and gain.\nIn particular, the IF is missing for the specified LO frequency and gain: LO={LO} Hz, IF={IF} Hz gain={gain}."
                )
            elif verbose_level == 1:
                print(
                    f"No correction parameters were found for the provided set of LO frequency, IF and gain.\nIn particular, the IF is missing for the specified LO frequency and gain: LO={LO} Hz, IF={IF} Hz gain={gain}."
                )
    else:
        if verbose_level == 2:
            raise Warning(
                f"No correction parameters were found for the provided set of LO frequency and gain: LO={LO} Hz, gain={gain}."
            )
        elif verbose_level == 1:
            print(
                f"No correction parameters were found for the provided set of LO frequency and gain: LO={LO} Hz, gain={gain}."
            )

    return param


def set_correction_parameters_to_opx(
    path_to_database: str,
    config: dict,
    element: str,
    LO: float,
    IF: float,
    gain: float,
    qm: QuantumMachine,
    job: Union[QmJob, RunningQmJob] = None,
    verbose_level: int = 2,
) -> dict:
    """Look for the correction parameters in the database, located at path_to_database, for the specified values of
    the Octave LO frequency, intermediate frequency and Octave gain and update the running job with `job.set_element_correction()`.
    If no job is specified, then the running will be taken using `qm.get_running_job()`.
    If no job is running, then the loaded mixer correction matrix is updated using `qm.set_mixer_correction()`.
    The correction parameters are returned in a dictionary of the form:
    ```{"offsets": {"I": I_offset, "Q": Q_offset}, "correction_matrix": correction}```.

    If the correction parameters are not found in the database:
      * verbose_level=2 will raise a warning which will block the execution.
      * verbose_level=1 will print a warning which will not block the execution and correction settings will be unchanged.
      * verbose_level=0 will not print anything and correction settings will be unchanged.
    :param path_to_database: direct path to the calibration database (without '/calibration_db.json').
    :param config: the OPX config dictionary.
    :param element: the element to get the correction parameters for.
    :param LO: the LO frequency in Hz.
    :param IF: the intermediate frequency in Hz.
    :param gain: the Octave gain.
    :param qm: the opened quantum machine.
    :param job: the running job. If None, then the running will be taken using `qm.get_running_job()`.
    :param verbose_level: set the type of messages printed in the Python console. Default is 2.
    :return: dictionary containing the 'I' and 'Q' offsets and the correction_matrix.
    """
    param = get_calibration_parameters_from_db(path_to_database, config, element, LO, IF, gain, verbose_level)
    if "I" in param["offsets"] and "Q" in param["offsets"]:
        qm.set_output_dc_offset_by_element(element, ("I", "Q"), (param["offsets"]["I"], param["offsets"]["Q"]))
    if job is None:
        job = qm.get_running_job()
        if job is not None:
            job.set_element_correction(element, param["correction_matrix"])
        else:
            mixer_name = qm.get_config()["elements"][element]["mixInputs"]["mixer"]
            lo_conf = qm.get_config()["elements"][element]["mixInputs"]["lo_frequency"]
            if_conf = qm.get_config()["elements"][element]["intermediate_frequency"]
            if IF != if_conf:
                print(
                    "WARNING: the intermediate frequency doesn't match the one defined in the config, which means that the loaded correction matrix may not be optimal."
                )
            qm.set_mixer_correction(mixer_name, if_conf, lo_conf, tuple(param["correction_matrix"]))
    return param


def get_correction_for_each_LO_and_IF(
    path_to_database: str,
    config: dict,
    element: str,
    gain: float,
    LO_list: Union[list, np.ndarray],
    IF_list: Union[list, np.ndarray],
    nb_of_updates: int,
    calibrate: bool = False,
    qm: QuantumMachine = None,
    calibration_params: AutoCalibrationParams = None,
):
    """Look in the calibration database for the calibration parameters corresponding to the provided set of LO
    frequencies, intermediate frequencies and gain.
    The intermediate frequencies considered here are only the ```nb_of_updates``` equally spaced frequencies from the
    provided ```IF_list```. For instance, if a list of 100 intermediate frequencies is provided, but nb_of_update is set
    to 10, then only 10 correction parameters will be returned for the returned equally spaced intermediate frequencies.

    The goal is to perform a wide frequency sweep (scan the LO frequency in Python and the IF in QUA) and update the
    mixer correction parameters for each LO frequency and a few intermediate frequencies, given by ```nb_of_updates```,
    in QUA.

    If the flag ```calibrate``` is set to True (the opened Quantum Machine needs to be provided), then the specified element will be calibrated at the given frequencies
    (all LO frequencies and only the ``nb_of_updates``` intermediate frequencies).
    Custom calibration parameters can be passed using the `calibration_params` dataclass. For instance:
    ```calibration_params = AutoCalibrationParams(); calibration_params.if_amplitude = 0.25```

    The function will return the list on intermediate frequencies at which the correction matrix will be updated in the
    program, the 'I' and 'Q' offsets and the four coefficients of the correction matrix with one element for each pair (LO, IF).

    :param path_to_database: direct path to the calibration database (without '/calibration_db.json').
    :param config: the OPX config dictionary.
    :param element: the element to get the correction parameters for.
    :param gain: the Octave gain.
    :param LO_list: list of the LO frequencies involved in the scan, in Hz.
    :param IF_list: list of the intermediate frequencies involved in the scan, in Hz.
    :param nb_of_updates: number of intermediate frequencies to calibrate and for which the program will update the correction pmatrix.
    :param calibrate: calibrate all the frequencies involved to the scan (LO and IF for the specified gain). Default is False.
    :param qm: the quantum machine object. Default is None.
    :param calibration_params: class containing the calibration parameters (if_amplitude, offset_frequency...). Default is None.
    :return: the list on intermediate frequencies at which the correction matrix will be updated in the
    program (size nb_of_updates), the 'I' and 'Q' offsets and the four coefficients of the correction matrix with one element for each pair
    (LO, IF) (size nb_of_updates*len(LO_list)): IFs, c00, c01, c10, c11, offset_I, offset_Q.
    """
    N = len(IF_list)
    IFs = [IF_list[i * N // nb_of_updates] for i in range(nb_of_updates)]
    c00 = []
    c01 = []
    c10 = []
    c11 = []
    offset_I = []
    offset_Q = []

    for lo in LO_list:
        if calibrate and qm is not None:
            qm.calibrate_element(element, {lo: tuple(IFs)}, params=calibration_params)
        elif calibrate and qm is None:
            raise Exception(
                "The opened Quantum Machine object must be provided if the flag ```calibrate``` is set to True."
            )

        for i in IFs:
            param = get_calibration_parameters_from_db(path_to_database, config, element, lo, i, gain)
            c00.append(param["correction_matrix"][0])
            c01.append(param["correction_matrix"][1])
            c10.append(param["correction_matrix"][2])
            c11.append(param["correction_matrix"][3])
            if i == IFs[0]:
                offset_I.append(param["offsets"]["I"])
                offset_Q.append(param["offsets"]["Q"])
    return IFs, c00, c01, c10, c11, offset_I, offset_Q


def octave_calibration_tool(
    qm: QuantumMachine,
    element: str,
    lo_frequencies: Union[int, float, list, np.ndarray],
    intermediate_frequencies: Union[int, float, list, np.ndarray],
):
    """Calibrate a given element for a list of LO and intermediate frequencies.
    :param qm: the quantum machine object.
    :param element: the element to calibrate.
    :param lo_frequencies: single value or list of LO frequencies to calibrate in Hz.
    :param intermediate_frequencies: single value or list of Intermediate frequencies to calibrate in Hz.
    """
    if not isinstance(lo_frequencies, (list, np.ndarray)):
        lo_frequencies = [lo_frequencies]

    if not isinstance(intermediate_frequencies, (list, np.ndarray)):
        intermediate_frequencies = [intermediate_frequencies]

    for lo in lo_frequencies:
        qm.calibrate_element(element, {lo: tuple(intermediate_frequencies)})



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import typing
from dash import dcc
import dash_bootstrap_components as dbc
from dash import html
from dash import Input, Output
import dash_dangerously_set_inner_html as innerHTML
from docutils.core import publish_doctree, publish_from_doctree
import inspect
import importlib
from .app import app
from .. import primitive_components, components

__all__ = [
    "editor_of_quantum_machine_elements",
    "component_editor",
]


def editor_of_quantum_machine_elements():

    qm_list = []
    component_libraries = ["qualang_tools.config.components"]

    for library in component_libraries:
        for name, obj in inspect.getmembers(importlib.import_module(library)):
            if inspect.isclass(obj) and obj.__module__ == library:
                qm_list.append({"label": f"{name}", "value": f"{library}.{name}"})
    qm_select = dcc.Dropdown(
        id="qm-component-select",
        options=qm_list,
        placeholder="Add new...",
        searchable=True,
        style={"font-size": "1.2rem"},
    )

    return dbc.Row(
        [
            dbc.Col(qm_select, width=3),
            dbc.Col("", id="component-gui", width=6),
            dbc.Col("", id="component-doc", width=3),
        ],
        style={"margin-top": "2rem"},
    )


@app.callback(
    [Output("component-gui", "children"), Output("component-doc", "children")],
    [Input("qm-component-select", "value")],
    prevent_initial_call=True,
)
def get_component_gui(selected_component):
    return component_editor(selected_component)


def component_editor(selected_component, inputs=None):
    if selected_component is None:
        return "", ""

    component, component_class = get_component(selected_component)

    docstring = get_docstring(component_class)

    arugments = inspect.getfullargspec(component_class.__init__)

    argument_names = arugments[0][1:]
    argument_types = arugments[6]
    component_gui = []

    import config_edits

    importlib.reload(config_edits)
    setup = config_edits.setup

    input_index = [0, 0]
    python_command = f"{component_class.__name__}("

    if inputs is None:
        inputs = [None, None]

    python_command, component_gui, input_index = class_editor_and_gui(
        component_class.__init__,
        setup,
        component_gui,
        python_command,
        inputs,
        input_index,
        input_types=["component-input", "port-list"],
    )

    # objects that we can add through add method
    additional_add_docstring = ""
    if hasattr(component_class, "add") and callable(component_class.add) and inputs[0] is None and inputs[1] is None:

        arugments = inspect.getfullargspec(component_class.add)
        argument_names = arugments[0][1:]
        argument_types = arugments[6]

        possible_add_on_objects_gui = []
        add_index = [0, 0]

        for index, name in enumerate(argument_names):

            possible_addons = [
                primitive_components.Operation,
                primitive_components.Weights,
                components.DigitalWaveform,
                components.Mixer,
                primitive_components.IntegrationWeights,
            ]
            for addon in possible_addons:
                # print(index," ",name, " ", argument_types[name])
                gui2 = []
                docs2 = ""
                option_name = ""
                if addon in typing.get_args(argument_types[name]):
                    add_class_name = f"{addon.__module__}.{addon.__qualname__}"
                    gui2.append(html.Small(add_class_name, className="text-muted"))
                    gui2.append(innerHTML.DangerouslySetInnerHTML(docs2))
                    forget, gui2, add_index = class_editor_and_gui(
                        addon,
                        setup,
                        gui2,
                        python_command,
                        inputs,
                        add_index,
                        input_types=[
                            "optional-class-input",
                            "optional-class-port-list",
                        ],
                    )
                    number_of_input_fields = len(gui2) - 2
                    gui2.append(
                        dbc.Input(
                            value=("%d" % number_of_input_fields),
                            id={
                                "type": "optional-class-fieldcount",
                                "index": add_index[0] + 1,
                            },
                            style={"display": "none"},
                        )
                    )
                    gui2.append(
                        dbc.Input(
                            value=(add_class_name),
                            id={
                                "type": "optional-class-name",
                                "index": add_index[0] + 1,
                            },
                            style={"display": "none"},
                        )
                    )
                    gui2.append(
                        dcc.Dropdown(
                            options=[],
                            value=[],
                            multi=True,
                            style={"display": "none"},
                            id={
                                "type": "optional-class-selection",
                                "index": add_index[0] + 1,
                            },
                        )
                    )

                    docs2 = get_docstring(addon)
                    option_name = addon.__name__
                    gui2.append(
                        dbc.Button(
                            "Add option to component",
                            color="secondary",
                            className="me-1",
                            id={"index": add_index[0] + 1, "type": "button-add-option"},
                        )
                    )
                    gui2.append(
                        html.Div(
                            "",
                            id={
                                "type": "optional-class-status",
                                "index": add_index[0] + 1,
                            },
                        )
                    )
                    possible_add_on_objects_gui.append(dbc.AccordionItem(gui2, title=option_name))

        component_gui.append(
            dbc.InputGroup(
                [
                    dbc.InputGroupText([html.I(className="bi bi-stack me-2"), "Add"]),
                    dcc.Dropdown(
                        options=[],
                        value=[],
                        multi=True,
                        placeholder="add optional objects using the menu below",
                        style={"width": "100%"},
                        id="component-input-add",
                    ),
                ],
                className="mb-3",
            )
        )
        if component_class.add.__doc__ is not None:
            additional_add_docstring = "Add: " + component_class.add.__doc__

        component_gui.append(
            dbc.Accordion(
                possible_add_on_objects_gui,
                start_collapsed=True,
            )
        )
    else:
        component_gui.append(
            dcc.Dropdown(
                options=[],
                value=[],
                multi=True,
                placeholder="add optional objects using the menu below",
                style={"display": "none"},
                id="component-input-add",
            )
        )

    if inputs[0] is not None or inputs[1] is not None:
        python_command += ")"
        return python_command

    component_gui.append(
        dbc.Button(
            "Add component",
            color="primary",
            className="me-1",
            id="button-add-component",
        )
    )
    component_gui.append(html.Div("", id="component-status"))
    docs = [
        html.H4(component),
        html.Small(selected_component, className="text-muted"),
        innerHTML.DangerouslySetInnerHTML(docstring),
        innerHTML.DangerouslySetInnerHTML(additional_add_docstring),
    ]

    return component_gui, docs


def one_or_single_type_of(some_type):
    return typing.Union[some_type, typing.List[some_type]]


def class_editor_and_gui(
    selected_class,
    setup,
    component_gui,
    python_command,
    inputs,
    input_index,
    input_types,
):
    arugments = inspect.getfullargspec(selected_class)
    # print(arugments)
    argument_names = arugments[0][1:]
    argument_types = arugments[6]

    for index_arg, name in enumerate(argument_names):
        # print(index_arg," ",name, " ",argument_types[name])
        if argument_types[name] == str:
            if inputs[0] is not None:
                value = inputs[0][input_index[0]].replace("'", "")
                python_command += (",\n  " if (index_arg > 0) else "") + f"'{value}'"
            else:
                component_gui.append(
                    dbc.InputGroup(
                        [
                            dbc.InputGroupText(name),
                            dbc.Input(
                                placeholder="string",
                                id={
                                    "type": input_types[0],
                                    "index": input_index[0],
                                },
                            ),
                        ],
                        className="mb-3",
                    )
                )
            input_index[0] += 1
        elif argument_types[name] == float:
            if inputs[0] is not None:
                value = float(inputs[0][input_index[0]])
                python_command += (",\n  " if (input_index[1] > 0 or input_index[0] > 0) else "") + f"{value}"
            else:
                component_gui.append(
                    dbc.InputGroup(
                        [
                            dbc.InputGroupText(name),
                            dbc.Input(
                                placeholder="float",
                                type="number",
                                id={
                                    "type": input_types[0],
                                    "index": input_index[0],
                                },
                            ),
                        ],
                        className="mb-3",
                    )
                )
            input_index[0] += 1
        elif argument_types[name] == int:
            if inputs[0] is not None:
                value = int(inputs[0][input_index[0]])
                python_command += (",\n  " if (input_index[1] > 0 or input_index[0] > 0) else "") + f"{value}"
            else:
                component_gui.append(
                    dbc.InputGroup(
                        [
                            dbc.InputGroupText(name),
                            dbc.Input(
                                placeholder="integer",
                                type="number",
                                step=1,
                                id={
                                    "type": input_types[0],
                                    "index": input_index[0],
                                },
                            ),
                        ],
                        className="mb-3",
                    )
                )
            input_index[0] += 1
        elif argument_types[name] == typing.List[float]:
            if inputs[0] is not None:
                python_command += (
                    ",\n  " if (input_index[1] > 0 or input_index[0] > 0) else ""
                ) + f"[{inputs[0][input_index[0]]}]"
            else:
                component_gui.append(
                    dbc.InputGroup(
                        [
                            dbc.InputGroupText(name),
                            dbc.Textarea(
                                placeholder="float, float, ...",
                                id={
                                    "type": input_types[0],
                                    "index": input_index[0],
                                },
                            ),
                        ],
                        className="mb-3",
                    ),
                )
            input_index[0] += 1
        elif argument_types[name] == typing.List[primitive_components.DigitalSample]:
            if inputs[0] is not None:
                python_command += (
                    ",\n  " if (input_index[1] > 0 or input_index[0] > 0) else ""
                ) + f"[{inputs[0][input_index[0]]}]"
            else:
                component_gui.append(
                    dbc.InputGroup(
                        [
                            dbc.InputGroupText(name),
                            dbc.Textarea(
                                placeholder="(0 or 1, duration in ns), (... , ...), ...",
                                id={
                                    "type": input_types[0],
                                    "index": input_index[0],
                                },
                            ),
                        ],
                        className="mb-3",
                    ),
                )
            input_index[0] += 1
        elif argument_types[name] in [
            primitive_components.AnalogOutputPort,
            primitive_components.AnalogInputPort,
            primitive_components.DigitalOutputPort,
            primitive_components.DigitalInputPort,
        ]:
            controllers = setup.find_instances(components.Controller)
            controllerList = []

            if inputs[0] is not None:
                controller_id = int(inputs[0][input_index[0]])
                port_id = int(inputs[0][input_index[0] + 1])
                command = f"setup.find_instances(Controller)[{controller_id}]"
                if argument_types[name] == primitive_components.AnalogOutputPort:
                    command += f".analog_output({port_id})"
                elif argument_types[name] == primitive_components.AnalogInputPort:
                    command += f".analog_input({port_id})"
                elif argument_types[name] == primitive_components.DigitalOutputPort:
                    command += f".digital_output({port_id})"
                elif argument_types[name] == primitive_components.DigitalInputPort:
                    command += f".digital_input({port_id})"
                python_command += (",\n" if (input_index[0] > 0 or input_index[1] > 0) else "") + command
            else:
                for index, c in enumerate(controllers):
                    controllerList.append({"label": c.name, "value": index})

                component_gui.append(
                    dbc.InputGroup(
                        [
                            dbc.InputGroupText([html.I(className="bi bi-diagram-3-fill me-2"), name]),
                            dcc.Dropdown(
                                options=controllerList,
                                placeholder="choose controller",
                                style={"min-width": "50%"},
                                id={
                                    "type": input_types[0],
                                    "index": input_index[0],
                                },
                            ),
                            dbc.InputGroupText("port_id"),
                            dbc.Input(
                                placeholder="integer",
                                type="number",
                                step=1,
                                id={
                                    "type": input_types[0],
                                    "index": input_index[0] + 1,
                                },
                            ),
                        ],
                        className="mb-3",
                    ),
                )
            input_index[0] += 2
        elif argument_types[name] in [
            typing.List[primitive_components.AnalogOutputPort],
            typing.List[primitive_components.AnalogInputPort],
            typing.List[primitive_components.DigitalInputPort],
            typing.List[primitive_components.DigitalOutputPort],
        ]:
            controllers = setup.find_instances(components.Controller)
            controllerList = []

            if inputs[0] is not None:
                list = "["

                for index, port in enumerate(inputs[1][input_index[1]]):
                    if index > 0:
                        list += ",\n  "
                    port = port.split(",")
                    controller_id = int(port[0])
                    port_id = int(port[1])
                    if argument_types[name] == typing.List[primitive_components.AnalogOutputPort]:
                        list += f"setup.find_instances(Controller)[{controller_id}].analog_output({port_id})"
                    elif argument_types[name] == typing.List[primitive_components.AnalogInputPort]:
                        list += f"setup.find_instances(Controller)[{controller_id}].analog_input({port_id})"
                    elif argument_types[name] == typing.List[primitive_components.DigitalOutputPort]:
                        list += f"setup.find_instances(Controller)[{controller_id}].digital_output({port_id})"
                    else:
                        # it is (argument_types[name] == typing.List[quantum_machine.primitive_components.DigitalInputPort])
                        list += f"setup.find_instances(Controller)[{controller_id}].digital_input({port_id})"

                list += "]"
                python_command += (",\n  " if (input_index[1] > 0 or input_index[0] > 0) else "") + list
            else:
                for index, c in enumerate(controllers):
                    controllerList.append({"label": c.name, "value": "%s, %d" % (c.name, index)})

                component_gui.append(
                    dbc.InputGroup(
                        [
                            dbc.InputGroupText("List of " + name + ""),
                            dcc.Dropdown(
                                options=[],
                                value=[],
                                multi=True,
                                placeholder="Add to list using form below",
                                style={"width": "100%"},
                                id={
                                    "type": input_types[1],
                                    "index": input_index[1],
                                },
                            ),
                        ]
                    )
                )

                # note: input_index[1] here doesn't change as these fields are not
                # returned from the final form (they are just in subform to help)
                # selection
                component_gui.append(
                    dbc.InputGroup(
                        [
                            dbc.InputGroupText(name[:-1]),
                            dcc.Dropdown(
                                options=controllerList,
                                placeholder="choose controller",
                                style={"min-width": "30%"},
                                id={
                                    "type": "input-controller",
                                    "index": input_index[1],
                                },
                            ),
                            dbc.InputGroupText("port_id"),
                            dbc.Input(
                                placeholder="integer",
                                type="number",
                                step=1,
                                id={
                                    "type": "input-port",
                                    "index": input_index[1],
                                },
                                style={"min-width": "10%"},
                            ),
                            dbc.Button(
                                "Add to list",
                                id={
                                    "type": "port-button",
                                    "index": input_index[1],
                                },
                            ),
                        ],
                        className="mb-3",
                    ),
                )

            input_index[1] += 1
        elif argument_types[name] == typing.List[primitive_components.Waveform]:
            waveforms = setup.find_instances(primitive_components.Waveform)
            waveformList = []
            if inputs[0] is not None:
                list = "["

                for index, component_index in enumerate(inputs[0][input_index[0]]):
                    if index > 0:
                        list += ",\n  "
                    list += f"setup.find_instances(Waveform)[{component_index}]"

                list += "]"
                python_command += (",\n  " if (input_index[1] > 0 or input_index[0] > 0) else "") + list
            else:
                for index, c in enumerate(waveforms):
                    waveformList.append({"label": c.name, "value": index})

                component_gui.append(
                    dbc.InputGroup(
                        [
                            dbc.InputGroupText("List of " + name + ""),
                            dcc.Dropdown(
                                options=waveformList,
                                value=[],
                                multi=True,
                                placeholder="Click and select available waveforms",
                                style={"width": "100%"},
                                id={
                                    "type": input_types[0],
                                    "index": input_index[0],
                                },
                            ),
                        ],
                        className="mb-3",
                    )
                )
            input_index[0] += 1
        elif argument_types[name] == primitive_components.Pulse:
            if inputs[0] is not None:
                python_command += (
                    ",\n" if (input_index[0] > 0 or input_index[1] > 0) else ""
                ) + f"setup.find_instances(Pulse)[{inputs[0][input_index[0]]}]"
            else:
                pulses = setup.find_instances(primitive_components.Pulse)
                pulseList = []
                for index, c in enumerate(pulses):
                    pulseList.append({"label": c.name, "value": index})

                component_gui.append(
                    dbc.InputGroup(
                        [
                            dbc.InputGroupText("List of " + name + ""),
                            dcc.Dropdown(
                                options=pulseList,
                                value=[],
                                multi=False,
                                placeholder="Select available pulse",
                                style={"width": "100%"},
                                id={
                                    "type": input_types[0],
                                    "index": input_index[0] + 1,
                                },
                            ),
                        ],
                        className="mb-3",
                    )
                )
            input_index[0] += 1
        elif argument_types[name] == typing.List[components.Element]:
            elements = setup.find_instances(components.Element)
            elementList = []
            if inputs[0] is not None:
                list = "["

                for index, component_index in enumerate(inputs[0][input_index[0]]):
                    if index > 0:
                        list += ",\n  "
                    list += f"setup.find_instances(Element)[{component_index}]"

                list += "]"
                python_command += (",\n  " if (input_index[1] > 0 or input_index[0] > 0) else "") + list
            else:
                for index, c in enumerate(elements):
                    elementList.append({"label": c.name, "value": index})

                component_gui.append(
                    dbc.InputGroup(
                        [
                            dbc.InputGroupText("List of " + name + ""),
                            dcc.Dropdown(
                                options=elementList,
                                value=[],
                                multi=True,
                                placeholder="Click and select available elements",
                                style={"width": "100%"},
                                id={
                                    "type": input_types[0],
                                    "index": input_index[0],
                                },
                            ),
                        ],
                        className="mb-3",
                    )
                )
            input_index[0] += 1
        elif argument_types[name] == primitive_components.IntegrationWeights:
            if inputs[0] is not None:

                selection = inputs[0][input_index[0]].split(",")[0]
                index = inputs[0][input_index[0]].split(",")[1]

                if selection == 1:
                    python_command += (
                        ",\n  " if (input_index[1] > 0 or input_index[0] > 0) else ""
                    ) + f"setup.find_instances(ConstantIntegrationWeights)[{index}]"
                else:
                    # selection == 2
                    python_command += (
                        ",\n  " if (input_index[1] > 0 or input_index[0] > 0) else ""
                    ) + f"setup.find_instances(ArbitraryIntegrationWeights)[{index}]"
            else:
                options1 = setup.find_instances(components.ConstantIntegrationWeights)
                options2 = setup.find_instances(components.ArbitraryIntegrationWeights)
                weights_list = []
                for index, c in enumerate(options1):
                    weights_list.append(
                        {
                            "label": "ConstantIntegrationWeights : " + c.name,
                            "value": f"1,{index}",
                        }
                    )
                for index, c in enumerate(options2):
                    weights_list.append(
                        {
                            "label": "ArbitraryIntegrationWeights : " + c.name,
                            "value": f"2,{index}",
                        }
                    )

                component_gui.append(
                    dbc.InputGroup(
                        [
                            dbc.InputGroupText("List of " + name + ""),
                            dcc.Dropdown(
                                options=weights_list,
                                value=[],
                                multi=False,
                                placeholder="Select integration weights",
                                style={"width": "100%"},
                                id={
                                    "type": input_types[0],
                                    "index": input_index[0],
                                },
                            ),
                        ],
                        className="mb-3",
                    )
                )
            input_index[0] += 1
        elif argument_types[name] == primitive_components.Matrix2x2:
            if inputs[0] is not None:
                a = float(inputs[0][input_index[0]])
                b = float(inputs[0][input_index[0] + 1])
                c = float(inputs[0][input_index[0] + 2])
                d = float(inputs[0][input_index[0] + 3])
                python_command += (
                    ",\n  " if (input_index[1] > 0 or input_index[0] > 0) else ""
                ) + f"Matrix2x2([[{a},{b}],[{c},{d}]])"
            else:
                component_gui.append(
                    dbc.InputGroup(
                        [
                            dbc.InputGroupText(name + " = [ ["),
                            dbc.Input(
                                placeholder="float",
                                type="number",
                                id={
                                    "type": input_types[0],
                                    "index": input_index[0],
                                },
                            ),
                            dbc.InputGroupText(" , "),
                            dbc.Input(
                                placeholder="float",
                                type="number",
                                id={
                                    "type": input_types[0],
                                    "index": input_index[0] + 1,
                                },
                            ),
                            dbc.InputGroupText("], ["),
                            dbc.Input(
                                placeholder="float",
                                type="number",
                                id={
                                    "type": input_types[0],
                                    "index": input_index[0] + 2,
                                },
                            ),
                            dbc.InputGroupText(" , "),
                            dbc.Input(
                                placeholder="float",
                                type="number",
                                id={
                                    "type": input_types[0],
                                    "index": input_index[0] + 3,
                                },
                            ),
                            dbc.InputGroupText("] ]"),
                        ],
                        className="mb-3",
                    ),
                )
            input_index[0] += 4
        else:
            print("Unknown data type")
            print(argument_types[name])

    return python_command, component_gui, input_index


def get_docstring(selected_class):
    if selected_class.__init__ is not None and selected_class.__init__.__doc__ is not None:
        docstring = publish_doctree(selected_class.__init__.__doc__)
        docstring = publish_from_doctree(docstring, writer_name="html").decode()
    else:
        docstring = ""
    return docstring


def get_component(selected_component):
    component_class = None
    if isinstance(selected_component, str):
        component = selected_component.split(".")[-1]
        library = inspect.getmembers(importlib.import_module(selected_component[0 : -len(component) - 1]))

        for name, obj in library:
            if name == component:
                component_class = obj
    else:
        component_class = selected_component
    return component, component_class



from .instruments import Instruments
from .connectivity import Connectivity
from .wirer.wirer import allocate_wiring
from .visualizer.web_visualizer import visualize
from .wirer.channel_specs_interface import (
    mw_fem_spec,
    lf_fem_spec,
    lf_fem_iq_spec,
    lf_fem_iq_octave_spec,
    opx_spec,
    opx_iq_spec,
    opx_iq_octave_spec,
    octave_spec,
)
from .wirer.wirer_exceptions import NotEnoughPulsersException, NotEnoughChannelsException

__all__ = [
    "Instruments",
    "Connectivity",
    "allocate_wiring",
    "visualize",
    "NotEnoughPulsersException",
    "NotEnoughChannelsException",
    "mw_fem_spec",
    "lf_fem_spec",
    "lf_fem_iq_spec",
    "lf_fem_iq_octave_spec",
    "opx_spec",
    "opx_iq_spec",
    "opx_iq_octave_spec",
    "octave_spec",
]



from pathlib import Path

from .data_processor import DataProcessor
from .helpers import (
    copy_nested_dict,
    iterate_nested_dict_with_lists,
    update_nested_dict,
)


class PlotlyGraphSaver(DataProcessor):
    file_format: str = "html"
    nested_separator: str = "."

    def __init__(self, file_format=None):
        if file_format is not None:
            self.file_format = file_format
        self.figures = {}

    @property
    def file_suffix(self) -> str:
        suffixes = {"html": ".html", "json": ".json"}
        return suffixes.get(self.file_format.lower(), ".html")

    def process(self, data):
        import plotly.graph_objs as go

        self.figures = {}
        processed_data = copy_nested_dict(data)

        for keys, val in iterate_nested_dict_with_lists(data):
            if not isinstance(val, go.Figure):
                continue

            path = Path(self.nested_separator.join(keys))
            self.figures[path] = val
            update_nested_dict(processed_data, keys, f"./{path}{self.file_suffix}")
        return processed_data

    def post_process(self, data_folder: Path):
        for path, fig in self.figures.items():
            file_path = data_folder / f"{path}{self.file_suffix}"
            file_path.parent.mkdir(parents=True, exist_ok=True)
            if self.file_format.lower() == "html":
                fig.write_html(str(file_path))
            elif self.file_format.lower() == "json":
                fig.write_json(str(file_path))
            else:
                raise NotImplementedError(f"File format {self.file_format} is not supported for Plotly graphs")



"""calling function libraries"""

from qm.qua import *
from qm import QuantumMachinesManager
import numpy as np
import matplotlib.pyplot as plt


class VNA:
    """
    Configures the OPX as a vector network analyzer (VNA) allowing to quickly perform basic VNA measurements:
        - S-parameter (S11 and S12)
        - Spectrum analyzer
    """

    def __init__(self, config: dict, element: str, operation: str, Q: int = None):
        """
        Gets a QUA configuration file and configure the OPX to perform VNA measurements.

        :param config: configuration file
        :param element: name of the element to be measured (defined in the configuration)
        :param operation: name of the measurement operation associated to the element (defined in the configuration)
        :param Q: value of the resonator quality factor for estimating the cavity relaxation time (optional).
        """
        # Check inputs errors
        if not (element in config["elements"]):
            raise KeyError(f"element {element} is not in the config")
        elif not (operation in config["elements"][element]["operations"]):
            raise KeyError(f"Operation {operation} is not in the config")

        self.qmm = None
        self.config = config
        self.element = element
        self.operation = operation
        self.inputs = [
            config["elements"][element]["outputs"][out][1] for out in config["elements"][element]["outputs"].keys()
        ]

        if Q is not None:
            self.relax_time = int(Q / (2 * np.pi * config["elements"][element]["mixInputs"]["lo_frequency"] * 1e-9) / 4)
        else:
            self.relax_time = 4
        self.measurements = []
        self.calibration = {}
        self.meas_list = []
        self.results = {}

    def __str__(self):
        s = f"\nVNA mode initialized for element '{self.element}'"
        s += f" and operation '{self.operation}'"
        s += f" with {len(self.measurements)} measurements:\n"
        for M in self.measurements:
            s += str(M) + "\n"
        return s

    def _add_S_measurement(self, measurement: str, outputs, int_weights):
        """
        Add measurement outputs and integration weights and initialize self.results.

        :param measurement: name of the measurement type ('S11' or 'S21')
        :param outputs: name of the output port of the element (defined in the configuration)
        :param int_weights:  name of the integration weights used for integration
        :return: None
        """
        self.measurements.append(
            {
                "measurement": measurement,
                "integration_weights": int_weights,
                "outputs": outputs,
            }
        )
        self.meas_list.append(measurement)
        self.results[measurement] = {"f": [], "S": [], "phase": []}

    def add_envelope_detector(self, measurement: str, outputs: str, int_weights: str):
        """
        Adds an S-measurement down-converted with an envelope detector.

        :param measurement: name of the measurement type ('S11' or 'S21')
        :param outputs: name of the output port of the element (defined in the configuration). For instance outputs="out1"
        :param int_weights: name of the integration weights used for integration. For instance int_weights="Wc"
        :return: None
        """
        # Check inputs errors
        if not (outputs in self.config["elements"][self.element]["outputs"]):
            raise KeyError(f"Output '{outputs}' is not in the config")
        elif not (
            int_weights
            in self.config["pulses"][self.config["elements"][self.element]["operations"][self.operation]][
                "integration_weights"
            ]
        ):
            raise KeyError(f"Integration weights '{int_weights}' is not in the config")
        elif not (measurement in ["S11", "S21"]):
            raise KeyError(f"Measurement '{measurement}' not implemented yet, must be in ['S11', 'S21']")

        self._add_S_measurement(measurement, outputs, int_weights)
        self.measurements[-1]["down_converter"] = "ED"

    def add_IR_mixer(self, measurement: str, outputs: str, int_weights: list):
        """
        Adds an S-measurement down-converted with an image rejection mixer.

        :param measurement: name of the measurement type ('S11' or 'S21')
        :param outputs: name of the output port of the element (defined in the configuration). for instance: outputs="out1"
        :param int_weights: names of the integration weights used for integration. For instance int_weights=["Wc", "Ws"]
        :return: None
        """
        # Check inputs errors
        if not (outputs in self.config["elements"][self.element]["outputs"]):
            raise KeyError(f"Output '{outputs}' is not in the config")
        elif not (
            all(
                [
                    iw
                    in self.config["pulses"][self.config["elements"][self.element]["operations"][self.operation]][
                        "integration_weights"
                    ]
                    for iw in int_weights
                ]
            )
        ):
            raise KeyError(f"Integration weights '{int_weights}' are not in the config")
        elif not (measurement in ["S11", "S21"]):
            raise KeyError(f"Measurement '{measurement}' not implemented yet, must be in ['S11', 'S21']")

        self._add_S_measurement(measurement, outputs, int_weights)
        self.measurements[-1]["down_converter"] = "IR"

    def add_IQ_mixer(self, measurement: str, outputs: list, int_weights: list):
        """
        Adds an S-measurement down-converted with an IQ mixer.

        :param measurement: name of the measurement type ('S11' or 'S21')
        :param outputs: name of the output ports of the element (defined in the configuration). For instance out=["out1", "out2"]
        :param int_weights: names of the integration weights used for integration. for instance int_weights=[["Wc", "Ws"], ["-Ws", "Wc"]]
        :return: None
        """
        # Check inputs errors
        if len(outputs) != 2:
            raise TypeError(f"Output '{outputs}' must be a list of two  outputs.")
        if np.array(int_weights).shape != (2, 2):
            raise TypeError(f"Output '{int_weights}' must be a 2x2 list.")
        if not (all([out in self.config["elements"][self.element]["outputs"] for out in outputs])):
            raise KeyError(f"Integration weights '{outputs}' are not in the config")
        elif not (
            all(
                [
                    iw
                    in self.config["pulses"][self.config["elements"][self.element]["operations"][self.operation]][
                        "integration_weights"
                    ]
                    for iw in np.array(int_weights).reshape(1, 4)[0]
                ]
            )
        ):
            raise KeyError(f"Integration weights '{int_weights}' are not in the config")
        elif not (measurement in ["S11", "S21"]):
            raise KeyError(f"Measurement '{measurement}' not implemented yet, must be in ['S11', 'S21']")

        self._add_S_measurement(measurement, outputs, int_weights)
        self.measurements[-1]["down_converter"] = "IQ"

    def add_spectrum_analyzer(self, bandwidth: int):
        """
        Add the spectrum analyzer measurement which will store its results under self.results["SA"]:
            - 'raw': raw_adc data [V]
            - 'fft': fft data [V]
            - 'f': frequency [Hz]

        :param bandwidth: Spectrum analyzer bandwidth [Hz]
        :return: None
        """
        self.meas_list.append("spectrum_analyzer")
        self.calibration["SA"] = {"pulse_length": int(2e9 / bandwidth)}
        self.results["SA"] = {"f": [], "raw": [], "fft": []}

    def _define_spectrum_analyzer(self, n_avg: int):
        """
        Define the spectrum analyzer QUA program

        :param n_avg: number of iterations for averaging.
        :return: Corresponding QUA program
        """
        with program() as Prog_spectro:
            n = declare(int)
            adc_st = declare_stream(adc_trace=True)

            with for_(n, 0, n < n_avg, n + 1):
                measure(self.operation, self.element, adc_st)

            with stream_processing():
                if 1 in self.inputs:
                    adc_st.input1().average().save("adc1_raw")
                    adc_st.input1().fft(output="abs").average().save("adc1_fft")
                if 2 in self.inputs:
                    adc_st.input2().average().save("adc2_raw")
                    adc_st.input2().fft(output="abs").average().save("adc2_fft")

        return Prog_spectro

    def _run_spectrum_analyzer(self, qm, n_avg: int):
        """
        Record raw adc and corresponding FFT.

        :param qm: opened quantum machine
        :param n_avg: number of averaged iterations
        :return: None
        """

        # Execute spectrum_analyzer calibration program
        job = qm.execute(self._define_spectrum_analyzer(n_avg))
        # Get results
        res_handles = job.result_handles
        res_handles.wait_for_all_values()
        fft = []
        raw = []
        for inp in self.inputs:
            # fft data
            # When doing fft, the frequency goes from -1/(2*dt) to 1/(2*dt) in jumps of  1/(maxT).
            # In our case, dt = 1ns and maxT = pulse_len.
            # In addition, the resulting fft data is shifted such that the order is [0:1/(2*dt), -1/(2*dt):0].
            # Here, instead of shifting the data, we just look only at the positive part.
            fft.append(res_handles.get(f"adc{inp}_fft").fetch_all())
            # raw data
            raw.append(res_handles.get(f"adc{inp}_raw").fetch_all())
        # Reformat fft and raw data
        fft = np.array(fft)
        raw = np.array(raw)
        # Store data in self.results
        self.results["SA"]["raw"] = raw / 4096
        self.results["SA"]["fft"] = fft / 4096 * (2 / self.calibration["SA"]["pulse_length"])
        self.results["SA"]["f"] = np.arange(0, 0.5 * 1e9, 1e9 / self.calibration["SA"]["pulse_length"])

    def _define_single_measurement(self, sequence: dict, freq_sweep: dict, n_avg: int):
        """
        Define a single S-measurement as a QUA program

        :param sequence: single measurement sequence
        :param freq_sweep: dictionary containing the frequency sweep information (fmin, fmax, step)
        :param n_avg: number of averaged iterations
        :return: corresponding QUA program
        """
        n_steps = (freq_sweep["f_max"] - freq_sweep["f_min"]) / freq_sweep["step"] + 1

        with program() as Prog_single:
            n = declare(int)
            f = declare(int)
            I = declare(fixed)
            Q = declare(fixed)
            I_st = declare_stream()
            Q_st = declare_stream()
            f_st = declare_stream()

            # Repeat the sequence n_avg times for averaging purposes
            with for_(n, 0, n < n_avg, n + 1):
                # Frequency sweep
                with for_(
                    f,
                    freq_sweep["f_min"],
                    f < freq_sweep["f_max"] + freq_sweep["step"] / 2,
                    f + freq_sweep["step"],
                ):
                    # Wait 3 photons lifetime until the resonator relaxes to vacuum (clock cycles)
                    wait(3 * self.relax_time, self.element)
                    # Update the IF
                    update_frequency(self.element, f, units="Hz")
                    # Measure and demodulate the transmitted signal
                    if sequence["down_converter"] == "ED":
                        measure(
                            self.operation,
                            self.element,
                            None,
                            integration.full(sequence["integration_weights"], I, sequence["outputs"]),
                        )

                    elif sequence["down_converter"] == "IR":
                        measure(
                            self.operation,
                            self.element,
                            None,
                            demod.full(
                                sequence["integration_weights"][0],
                                I,
                                sequence["outputs"],
                            ),
                            demod.full(
                                sequence["integration_weights"][1],
                                Q,
                                sequence["outputs"],
                            ),
                        )

                    elif sequence["down_converter"] == "IQ":
                        measure(
                            self.operation,
                            self.element,
                            None,
                            dual_demod.full(
                                sequence["integration_weights"][0][0],
                                sequence["outputs"][0],
                                sequence["integration_weights"][0][1],
                                sequence["outputs"][1],
                                I,
                            ),
                            dual_demod.full(
                                sequence["integration_weights"][1][0],
                                sequence["outputs"][0],
                                sequence["integration_weights"][1][1],
                                sequence["outputs"][1],
                                Q,
                            ),
                        )
                    save(I, I_st)
                    save(Q, Q_st)
                    save(f, f_st)

            with stream_processing():
                I_st.buffer(n_steps).average().save("I")
                Q_st.buffer(n_steps).average().save("Q")
                f_st.buffer(n_steps).save("f")
        return Prog_single

    def _run_single_measurement(self, qm, sequence: dict, freq_sweep: dict, n_avg: int):
        """
        Execute the single_measurement QUA program

        :param qm: opened quantum machine
        :param sequence: single measurement sequence
        :param freq_sweep: dictionary containing the frequency sweep information (fmin, fmax, step)
        :param n_avg: number of averaged iterations
        :return: None
        """

        print(f"Run {sequence['measurement']} measurement...")
        # Execute the measurement program
        job = qm.execute(self._define_single_measurement(sequence, freq_sweep, n_avg))
        # Fetch results
        print("Fetching data...")
        res_handles = job.result_handles
        res_handles.wait_for_all_values()
        I = res_handles.get("I").fetch_all()
        Q = res_handles.get("Q").fetch_all()
        f = res_handles.get("f").fetch_all()

        self.results[sequence["measurement"]]["f"] = f
        if sequence["down_converter"] == "ED":
            self.results[sequence["measurement"]]["S"] = (
                np.abs(I)
                * 4096
                / (self.config["pulses"][self.config["elements"][self.element]["operations"][self.operation]]["length"])
            )
            self.results[sequence["measurement"]]["phase"] = 0 * np.abs(I)
        else:
            self.results[sequence["measurement"]]["S"] = np.sqrt(I**2 + Q**2)
            self.results[sequence["measurement"]]["phase"] = np.unwrap(np.arctan2(I, Q))

    def _define_dual_measurement(self, freq_sweep: dict, n_avg: int):
        """
        Define a dual S-measurement (simultaneous S11 and S21) as a QUA program

        :param freq_sweep: dictionary containing the frequency sweep information (fmin, fmax, step)
        :param n_avg: number of averaged iterations
        :return: corresponding QUA program
        """
        n_steps = (freq_sweep["f_max"] - freq_sweep["f_min"]) / freq_sweep["step"] + 1

        with program() as Prog_dual:
            n = declare(int)
            f = declare(int)
            I = declare(fixed)
            Q = declare(fixed)
            I2 = declare(fixed)
            Q2 = declare(fixed)
            I_st = declare_stream()
            Q_st = declare_stream()
            I2_st = declare_stream()
            Q2_st = declare_stream()
            f_st = declare_stream()

            # Repeat the sequence n_avg times for averaging purposes
            with for_(n, 0, n < n_avg, n + 1):
                # Frequency sweep
                with for_(
                    f,
                    freq_sweep["f_min"],
                    f < freq_sweep["f_max"] + freq_sweep["step"] / 2,
                    f + freq_sweep["step"],
                ):
                    # Wait 3 photons lifetime until the resonator relaxes to vacuum (clock cycles)
                    wait(3 * self.relax_time, self.element)
                    # Update the IF
                    update_frequency(self.element, f, units="Hz")
                    # Measure and demodulate the transmitted signal
                    if (
                        self.measurements[0]["down_converter"] == "ED"
                        and self.measurements[1]["down_converter"] == "ED"
                    ):
                        measure(
                            self.operation,
                            self.element,
                            None,
                            integration.full(
                                self.measurements[0]["integration_weights"],
                                I,
                                self.measurements[0]["outputs"],
                            ),
                            integration.full(
                                self.measurements[1]["integration_weights"],
                                I2,
                                self.measurements[1]["outputs"],
                            ),
                        )

                    elif (
                        self.measurements[0]["down_converter"] == "IR"
                        and self.measurements[1]["down_converter"] == "IR"
                    ):
                        measure(
                            self.operation,
                            self.element,
                            None,
                            demod.full(
                                self.measurements[0]["integration_weights"][0],
                                I,
                                self.measurements[0]["outputs"],
                            ),
                            demod.full(
                                self.measurements[0]["integration_weights"][1],
                                Q,
                                self.measurements[0]["outputs"],
                            ),
                            demod.full(
                                self.measurements[1]["integration_weights"][0],
                                I2,
                                self.measurements[1]["outputs"],
                            ),
                            demod.full(
                                self.measurements[1]["integration_weights"][1],
                                Q2,
                                self.measurements[1]["outputs"],
                            ),
                        )

                    elif (
                        self.measurements[0]["down_converter"] == "IQ"
                        and self.measurements[1]["down_converter"] == "IQ"
                    ):
                        measure(
                            self.operation,
                            self.element,
                            None,
                            dual_demod.full(
                                self.measurements[0]["integration_weights"][0][0],
                                self.measurements[0]["outputs"][0],
                                self.measurements[0]["integration_weights"][0][1],
                                self.measurements[0]["outputs"][1],
                                I,
                            ),
                            dual_demod.full(
                                self.measurements[0]["integration_weights"][1][0],
                                self.measurements[0]["outputs"][0],
                                self.measurements[0]["integration_weights"][1][1],
                                self.measurements[0]["outputs"][1],
                                Q,
                            ),
                            dual_demod.full(
                                self.measurements[1]["integration_weights"][0][0],
                                self.measurements[1]["outputs"][0],
                                self.measurements[1]["integration_weights"][0][1],
                                self.measurements[1]["outputs"][1],
                                I2,
                            ),
                            dual_demod.full(
                                self.measurements[1]["integration_weights"][1][0],
                                self.measurements[1]["outputs"][0],
                                self.measurements[1]["integration_weights"][1][1],
                                self.measurements[1]["outputs"][1],
                                Q2,
                            ),
                        )
                    else:
                        raise KeyError(
                            f"Dual measurement with '{self.measurements[0]['down_converter']}' and "
                            f"'{self.measurements[1]['down_converter']}' is not implemented yet."
                        )
                    save(I, I_st)
                    save(Q, Q_st)
                    save(I2, I2_st)
                    save(Q2, Q2_st)
                    save(f, f_st)

            with stream_processing():
                I_st.buffer(n_steps).average().save("I")
                Q_st.buffer(n_steps).average().save("Q")
                I2_st.buffer(n_steps).average().save("I2")
                Q2_st.buffer(n_steps).average().save("Q2")
                f_st.buffer(n_steps).save("f")
        return Prog_dual

    def _run_dual_measurement(self, qm, freq_sweep: dict, n_avg: int):
        """
        Execute the dual_measurement QUA program

        :param qm: opened quantum machine
        :param freq_sweep: dictionary containing the frequency sweep information (fmin, fmax, step)
        :param n_avg: number of averaged iterations
        :return: None
        """
        # Get defined program
        Prog = self._define_dual_measurement(freq_sweep, n_avg)
        print(
            f"Run dual measurements {self.measurements[0]['measurement']} and {self.measurements[1]['measurement']}..."
        )
        # Execute the measurement program
        job = qm.execute(Prog)

        # Fetch results
        print("Fetching data...")
        res_handles = job.result_handles
        res_handles.wait_for_all_values()

        f = res_handles.get("f").fetch_all()
        for i in range(2):
            if i == 1:
                I = res_handles.get("I").fetch_all()
                Q = res_handles.get("Q").fetch_all()
            else:
                I = res_handles.get("I2").fetch_all()
                Q = res_handles.get("Q2").fetch_all()
            self.results[self.measurements[i]["measurement"]]["f"] = f
            if self.measurements[i]["down_converter"] == "ED":
                self.results[self.measurements[i]["measurement"]]["S"] = np.abs(I)
                self.results[self.measurements[i]["measurement"]]["phase"] = 0 * np.abs(I)
            else:
                self.results[self.measurements[i]["measurement"]]["S"] = np.sqrt(I**2 + Q**2)
                self.results[self.measurements[i]["measurement"]]["phase"] = np.unwrap(np.arctan2(I, Q))

    def run_all(self, freq_sweep: dict = None, n_avg: int = 1, dual: bool = False):
        """
        Run all previously defined VNA measurements and store results in self.results.

        :param freq_sweep: dictionary containing the frequency sweep information. For instance freq_sweep = {"fmin": 10e6,"fmax": 150e6,"step": 0.01e6,}
        :param n_avg: number of averaged iterations
        :param dual: Boolean flag controlling the parallelism of the measurements. If two measurements were added, then dual=True
        will run them simultaneously, whereas if dual=False the measurements will be run sequentially.
        :return: None
        """
        print(self)
        # Open quantum machine manager
        self.qmm = QuantumMachinesManager()

        if "spectrum_analyzer" in self.meas_list:
            print("Spectrum analyzer...")
            # update measurement pulse duration to match bandwidth
            pulse = self.config["elements"][self.element]["operations"][self.operation]
            self.config["pulses"][pulse]["length"] = self.calibration["SA"]["pulse_length"]
            qm = self.qmm.open_qm(self.config)
            self._run_spectrum_analyzer(qm, n_avg)

        # Run each measurement sequentially or in parallel if dual
        qm = self.qmm.open_qm(self.config)
        if not dual:
            for seq in self.measurements:
                self._run_single_measurement(qm, seq, freq_sweep, n_avg)
        else:
            if len(self.measurements) != 2:
                raise KeyError("Dual run only works for two previously defined measurements")
            self._run_dual_measurement(qm, freq_sweep, n_avg)

    def plot_all(self):
        """
        Plot all the defined measurements.

        :return: None
        """

        for ss in self.results.keys():
            # Spectrum analyzer plot
            if ss == "SA":
                plt.figure(1001, figsize=(9, 8))
                for i in range(len(self.inputs)):
                    plt.subplot(211)
                    plt.plot(
                        np.arange(0, self.calibration["SA"]["pulse_length"], 1),
                        self.results["SA"]["raw"][i],
                        label=f"OPX input {self.inputs[i]}",
                    )
                    plt.subplot(212)
                    plt.plot(
                        self.results["SA"]["f"],
                        self.results["SA"]["fft"][i][: int(np.ceil(len(self.results["SA"]["fft"][i]) / 2))],
                        label=f"OPX input {self.inputs[i]}",
                    )
                plt.subplot(211)
                plt.title("VNA spectrum analyzer")
                plt.xlabel("Time [ns]")
                plt.ylabel("Raw ADC trace [V]")
                plt.legend()
                plt.subplot(212)
                plt.xlabel("Frequency [Hz]")
                plt.ylabel("FFT [V]")
                plt.legend()
            # S-parameter measurement
            else:
                plt.figure(1002, figsize=(9, 8))
                plt.subplot(211)
                plt.title(f"Spectroscopy of '{self.element}'")
                plt.plot(self.results[ss]["f"] / 1e6, self.results[ss]["S"], label=ss)
                plt.ylabel(r"$\sqrt{I^2+Q^2}$ [V]")
                plt.legend()
                plt.subplot(212)
                plt.plot(self.results[ss]["f"] / 1e6, self.results[ss]["phase"], label=ss)
                plt.ylabel("phase [rad]")
                plt.xlabel("Frequency [Hz]")
                plt.legend()
        plt.show()



import logging
from contextlib import contextmanager
from time import sleep, time
from typing import Generator

from qm import QuantumMachinesManager, QuantumMachine, QopCaps
from qm.logging_utils import set_logging_level

msg = "cannot be used because it isn't shareable in other QM."
msg_opx1000 = "Resources already locked"


@contextmanager
def qm_session(qmm: QuantumMachinesManager, config: dict, timeout: int = 100) -> Generator[QuantumMachine, None, None]:
    """
    This context manager allows a user to _try_ to
    open a quantum machine, and if it is not possible since its resources are currently in use,
    wait for them to free up. This is done by repeatedly polling the QM manager.

    Once they are freed, execution will start automatically.
    Once the context manager code block is exited, the QM will close automatically, freeing it up
    for another user.

    :param qmm: A QM manager from which to open QM
    :param config: a QUA config that will be supplied to `open_qm`
    :param timeout: time in seconds to wait for resources to free up before raising exception

    """
    if not timeout > 0:
        raise ValueError(f"{timeout=} must be positive")

    qm_log = logging.getLogger("qm.api.frontend_api")  # formerly "qm" & qm.api.frontend_api
    is_busy = True
    printed = False
    t0 = time()
    elapsed_time = 0
    while is_busy and elapsed_time < timeout:
        try:
            qm = qmm.open_qm(config, close_other_machines=False)

        except Exception as e:
            if (qmm.capabilities.supports(QopCaps.qop3) and msg_opx1000 in str(e)) or (
                ~qmm.capabilities.supports(QopCaps.qop3) and msg in str(e)
            ):
                if not printed:
                    qm_log.error(f"QOP is busy. Waiting for it to free up for {timeout}s...")
                    set_logging_level("CRITICAL")

                    printed = True
                sleep(0.2)
                elapsed_time = time() - t0
            else:
                raise Exception from e
        else:
            is_busy = False
            set_logging_level("INFO")
            qm_log.info("Opening QM")

    if is_busy and elapsed_time >= timeout:
        qm_log.warning(f"While waiting for QOP to free, reached timeout: {timeout}s")
        raise TimeoutError(f"While waiting for QOP to free, reached timeout: {timeout}s")
    try:
        yield qm

        if qmm.capabilities.supports(QopCaps.qop3):
            while qm.get_jobs(status=["Running"]):
                sleep(0.2)
        else:
            while qm.get_running_job() is not None:
                sleep(0.2)
    except KeyboardInterrupt:
        pass
    finally:
        qm_log.info("Closing QM")
        qm.close()



from typing import Dict, Optional

import qm.api
from packaging.version import Version
from qm import QuantumMachine
from qm.api.models.compiler import CompilerOptionArguments
from qm.jobs.running_qm_job import RunningQmJob
from qm.program import Program
from qm.simulate.interface import SimulationConfig
import qm

# TODO: Remove this if block when we drop support for qm < 1.2.3 (and keep only the import that is currently in the
#  else block)
qua_below_1_2_3 = Version(qm.__version__) < Version("1.2.3")
if qua_below_1_2_3:
    from qm.qua._dsl import _ProgramScope as _ProgramScope_qua
else:
    from qm.qua._scope_management._core_scopes import _ProgramScope as _ProgramScope_qua

__all__ = [
    "patch_qua_program_addons",
]


class _ProgramScope(_ProgramScope_qua):
    def __enter__(self) -> Program:
        program = super().__enter__()

        # Instantiate all addons
        program.addons = {name: addon() for name, addon in Program.addons.items()}

        # TODO Should we add exception handling?
        for addon in self._program.addons.values():
            addon.enter_program(program)

        return self._program

    def __exit__(self, exc_type, exc_val, exc_tb):
        for addon in self._program.addons.values():
            # TODO Should we add exception handling?
            addon.exit_program(exc_type, exc_val, exc_tb)

        exit_result = super().__exit__(exc_type, exc_val, exc_tb)

        return exit_result


QM_execute = QuantumMachine.execute


def QM_execute_patched(
    self,
    program: Program,
    duration_limit: int = 1000,
    data_limit: int = 20000,
    force_execution: int = False,
    dry_run: int = False,
    simulate: Optional[SimulationConfig] = None,
    compiler_options: Optional[CompilerOptionArguments] = None,
) -> RunningQmJob:
    return_val = QM_execute(
        self,
        program,
        duration_limit,
        data_limit,
        force_execution,
        dry_run,
        simulate,
        compiler_options,
    )

    for program_addon in program.addons.values():
        program_addon.execute_program(program=program, quantum_machine=self)

    return return_val


QM_execute_old_api = qm.api.v2.qm_api_old.QmApiWithDeprecations.execute


def QM_execute_patched_old_api(
    self,
    program: Program,
    duration_limit: int = 1000,
    data_limit: int = 20000,
    force_execution: int = False,
    dry_run: int = False,
    simulate: Optional[SimulationConfig] = None,
    compiler_options: Optional[CompilerOptionArguments] = None,
) -> RunningQmJob:
    return_val = QM_execute_old_api(
        self,
        program,
        duration_limit,
        data_limit,
        force_execution,
        dry_run,
        simulate,
        compiler_options=compiler_options,
    )

    for program_addon in program.addons.values():
        program_addon.execute_program(program=program, quantum_machine=self)

    return return_val


def patch_qua_program_addons():

    if hasattr(Program, "addons"):
        print("qm.program.Program already has 'addons' attribute, not patching")
    else:
        from qualang_tools.callable_from_qua import ProgramAddon

        Program.addons: Dict[str, ProgramAddon] = {}

    # TODO: Remove this if block when we drop support for qm < 1.2.3 (and keep only the import that is currently in the
    #  else block)
    if qua_below_1_2_3:
        import qm.qua._dsl

        if qm.qua._dsl._ProgramScope is _ProgramScope:
            print("qm.qua._dsl._ProgramScope already patched, not patching")
        else:
            qm.qua._dsl._ProgramScope = _ProgramScope
    else:
        # QUA >= 1.2.3: program() constructs _ProgramScope from qm.qua._dsl.scope_functions
        # To avoid class identity mismatches, patch the factory used by program() only.
        from qm.qua._dsl import scope_functions as _scope_functions

        if _scope_functions._ProgramScope is _ProgramScope:
            print("qm.qua._dsl.scope_functions._ProgramScope already patched, not patching")
        else:
            _scope_functions._ProgramScope = _ProgramScope

    # Patch QuantumMachine.execute without referencing the module name to avoid local scope confusion
    if QuantumMachine.execute is QM_execute_patched:
        print("qm.QuantumMachine.QuantumMachine.execute already patched, not patching")
    else:
        QuantumMachine.execute = QM_execute_patched

    # Patch old API execute by importing the class explicitly
    from qm.api.v2.qm_api_old import QmApiWithDeprecations as _QmApiWithDeprecations

    if _QmApiWithDeprecations.execute is QM_execute_patched_old_api:
        print("qm.QuantumMachine.QuantumMachine.execute already patched, not patching")
    else:
        _QmApiWithDeprecations.execute = QM_execute_patched_old_api



"""
The purpose of this module is to compile a sequence of wiring specifications
into potential channel specifications, then to allocate them to the first valid
combination of instrument channels.
"""

import copy
from typing import List

from .channel_specs import (
    ChannelSpecLfFemSingle,
    ChannelSpecOpxPlusSingle,
    ChannelSpecMwFemSingle,
    ChannelSpecLfFemBaseband,
    ChannelSpecOctave,
    ChannelSpecExternalMixer,
    ChannelSpecOpxPlusBaseband,
    ChannelSpecMwFemDigital,
    ChannelSpecLfFemDigital,
    ChannelSpecOctaveDigital,
    ChannelSpecOpxPlusDigital,
    ChannelSpecExternalMixerDigital,
)
from .wirer_assign_channels_to_spec import assign_channels_to_spec
from .wirer_exceptions import ConstraintsTooStrictException, NotEnoughChannelsException
from ..connectivity.channel_spec import ChannelSpec
from ..instruments import Instruments
from ..connectivity import Connectivity
from ..connectivity.wiring_spec import WiringSpec, WiringFrequency, WiringLineType


def allocate_wiring(
    connectivity: Connectivity,
    instruments: Instruments,
    block_used_channels: bool = True,
    clear_wiring_specifications: bool = True,
    observe_pulser_allocation: bool = False,
):
    """
    Allocates available instrument channels to quantum elements based on their wiring specifications.

    This function takes the wiring specifications from the provided `Connectivity` instance and attempts
    to allocate channels for each specification to the available instrument channels within the `Instruments` instance.

    The allocation is done in a sequence based on the line types specified in the `line_type_fill_order`. The function
    handles different frequency requirements, either DC or RF, and manages the reuse of available channels depending on
    the `block_used_channels` flag.

    Flow overview:
    1. Split the wiring specs by `line_type`, then separate each group into constrained and unconstrained specs.
    2. Allocate all constrained specs first (across all line types), then any constrained specs with untyped line types.
       This prevents unconstrained allocations from consuming scarce channels required by fixed constraints.
    3. Allocate remaining unconstrained specs in line-type order, followed by unconstrained untyped specs.
    4. If `clear_wiring_specifications` is `True`, the specifications are cleared after the allocation.
    5. If `block_used_channels` is `False`, channels newly used during allocation are returned to the available pool.

    Args:
        connectivity (Connectivity): An instance of the `Connectivity` class containing the wiring specifications to be allocated.
        instruments (Instruments): An instance of the `Instruments` class that manages available and used channels.
        block_used_channels (bool, optional): If `True`, prevents previously used channels from being returned to the available pool.
            Defaults to `True`.
        clear_wiring_specifications (bool, optional): If `True`, clears the list of wiring specifications in `connectivity`
            after allocation. Defaults to `True`.
        observe_pulser_allocation (bool, optional): If `True`, enforces pulser allocation constraints (16 pulsers per FEM,
            18 per OPX+). If `False`, ignores pulser limits and only respects physical channel limits.
            Defaults to `False`.

    Raises:
        ConstraintsTooStrictException: If the constraints for a wiring specification are too strict, preventing allocation.
        NotEnoughChannelsException: If there are not enough available channels to satisfy the wiring specification.
    """

    line_type_fill_order = [t for t in WiringLineType]

    specs = connectivity.specs

    used_channel_cache = copy.deepcopy(instruments.used_channels)

    # Always allocate constrained specs first across all line types to avoid
    # unconstrained wiring consuming scarce, fixed resources.
    typed_specs_by_line_type = {
        line_type: [spec for spec in specs if spec.line_type == line_type] for line_type in line_type_fill_order
    }
    constrained_specs_by_line_type = {
        line_type: [spec for spec in typed_specs_by_line_type[line_type] if spec.constraints]
        for line_type in line_type_fill_order
    }
    unconstrained_specs_by_line_type = {
        line_type: [spec for spec in typed_specs_by_line_type[line_type] if not spec.constraints]
        for line_type in line_type_fill_order
    }

    specs_with_untyped_lines = [spec for spec in specs if spec.line_type not in line_type_fill_order]

    # remove duplicates
    specs_with_untyped_lines = list(dict.fromkeys(specs_with_untyped_lines))
    constrained_untyped_specs = [spec for spec in specs_with_untyped_lines if spec.constraints]
    unconstrained_untyped_specs = [spec for spec in specs_with_untyped_lines if not spec.constraints]

    for line_type in line_type_fill_order:
        for spec in constrained_specs_by_line_type[line_type]:
            _allocate_wiring(spec, instruments, observe_pulser_allocation)
    for spec in constrained_untyped_specs:
        _allocate_wiring(spec, instruments, observe_pulser_allocation)

    for line_type in line_type_fill_order:
        for spec in unconstrained_specs_by_line_type[line_type]:
            _allocate_wiring(spec, instruments, observe_pulser_allocation)
    for spec in unconstrained_untyped_specs:
        _allocate_wiring(spec, instruments, observe_pulser_allocation)

    if clear_wiring_specifications:
        connectivity.specs = []

    if not block_used_channels:
        for channel_type, used_channels in instruments.used_channels.items():
            for i, used_channel in enumerate(reversed(used_channels)):
                if used_channel not in used_channel_cache:
                    instruments.available_channels.insert(0, used_channel)
                    instruments.used_channels.remove(used_channel)


def _allocate_wiring(spec: WiringSpec, instruments: Instruments, observe_pulser_allocation: bool = False):
    if spec.frequency == WiringFrequency.DC:
        allocate_dc_channels(spec, instruments, observe_pulser_allocation)

    elif spec.frequency == WiringFrequency.RF:
        allocate_rf_channels(spec, instruments, observe_pulser_allocation)

    elif spec.frequency == WiringFrequency.DO:
        allocate_do_channels(spec, instruments, observe_pulser_allocation)

    else:
        raise NotImplementedError()


def allocate_dc_channels(spec: WiringSpec, instruments: Instruments, observe_pulser_allocation: bool = False):
    """
    Try to allocate DC channels to an LF-FEM or OPX+ to satisfy the spec.
    """
    dc_specs = [
        # LF-FEM, Single analog output
        ChannelSpecLfFemSingle() & ChannelSpecLfFemDigital(),
        # OPX+, Single analog output
        ChannelSpecOpxPlusSingle() & ChannelSpecOpxPlusDigital(),
    ]

    allocate_channels(
        spec, dc_specs, instruments, same_con=True, same_slot=True, observe_pulser_allocation=observe_pulser_allocation
    )


def allocate_rf_channels(spec: WiringSpec, instruments: Instruments, observe_pulser_allocation: bool = False):
    """
    Try to allocate RF channels to a MW-FEM. If that doesn't work, look for a
    combination of LF-FEM I/Q and Octave channels, or OPX+ I/Q and Octave
    channels.
    """
    rf_specs = [
        # MW-FEM, Single RF output
        ChannelSpecMwFemSingle() & ChannelSpecMwFemDigital(),
        # LF-FEM I/Q output with Octave for upconversion
        ChannelSpecLfFemBaseband() & ChannelSpecLfFemDigital() & ChannelSpecOctave() & ChannelSpecOctaveDigital(),
        # LF-FEM I/Q output with External Mixer for upconversion
        ChannelSpecLfFemBaseband()
        & ChannelSpecLfFemDigital()
        & ChannelSpecExternalMixer()
        & ChannelSpecExternalMixerDigital(),
        # OPX+ I/Q output with Octave for upconversion
        ChannelSpecOpxPlusBaseband() & ChannelSpecOpxPlusDigital() & ChannelSpecOctave() & ChannelSpecOctaveDigital(),
        # OPX+ I/Q output with External Mixer for upconversion
        ChannelSpecOpxPlusBaseband()
        & ChannelSpecOpxPlusDigital()
        & ChannelSpecExternalMixer()
        & ChannelSpecExternalMixerDigital(),
    ]

    allocate_channels(
        spec, rf_specs, instruments, same_con=True, same_slot=True, observe_pulser_allocation=observe_pulser_allocation
    )


def allocate_do_channels(spec: WiringSpec, instruments: Instruments, observe_pulser_allocation: bool = False):
    """
    Try to allocate Digital Only (DO) channels to an LF- or MW-FEM or OPX+ to satisfy the spec.
    """
    do_specs = [
        # LF-FEM, Single digital output
        ChannelSpecLfFemDigital(),
        # MW-FEM, Single digital output
        ChannelSpecMwFemDigital(),
        # OPX+, Single digital output
        ChannelSpecOpxPlusDigital(),
    ]

    allocate_channels(
        spec, do_specs, instruments, same_con=True, same_slot=True, observe_pulser_allocation=observe_pulser_allocation
    )


def allocate_channels(
    wiring_spec: WiringSpec,
    channel_specs: List[ChannelSpec],
    instruments: Instruments,
    same_con: bool,
    same_slot: bool,
    observe_pulser_allocation: bool = False,
):
    mask_failures = 0
    for channel_spec in channel_specs:
        channel_spec = channel_spec.filter_by_wiring_spec(wiring_spec)
        if wiring_spec.constraints:
            constraints = wiring_spec.constraints.filter_by_wiring_spec(wiring_spec)
            if not channel_spec.apply_constraints(constraints):
                mask_failures += 1
                continue
        if assign_channels_to_spec(
            wiring_spec,
            instruments,
            channel_spec.channel_templates,
            same_con=same_con,
            same_slot=same_slot,
            observe_pulser_allocation=observe_pulser_allocation,
        ):
            return

    if mask_failures == len(channel_specs):
        raise ConstraintsTooStrictException(wiring_spec, wiring_spec.constraints)
    else:
        raise NotEnoughChannelsException(wiring_spec)



"""Tools for handling data folders.

This module provides functions for handling data folders, including extracting properties from data folders,
finding the latest data folder, generating relative folder names, and creating new data folders.
"""

from pathlib import Path
from typing import Dict, Union, Optional
import re
from datetime import datetime


__all__ = ["DEFAULT_FOLDER_PATTERN", "extract_data_folder_properties", "get_latest_data_folder", "create_data_folder"]


DEFAULT_FOLDER_PATTERN = "%Y-%m-%d/#{idx}_{name}_%H%M%S"


def _validate_datetime(datetime_str: str, datetime_format: str) -> bool:
    """Validate a datetime string with a given format.

    :param datetime_str: The datetime string to validate.
    :param datetime_format: The format of the datetime string.
    :return: True if the datetime string is valid, False otherwise.
    """
    try:
        datetime.strptime(datetime_str, datetime_format)
    except ValueError:
        return False
    return True


def extract_data_folder_properties(
    data_folder: Path, pattern: str = DEFAULT_FOLDER_PATTERN, root_data_folder: Path = None
) -> Optional[Dict[str, Union[str, int, Path]]]:
    """Extract properties from a data folder.

    :param data_folder: The data folder to extract properties from. Should be an absolute path.
    :param pattern: The pattern to extract the properties from, e.g. "#{idx}_{name}_%H%M%S".
    :param root_data_folder: The root data folder to extract the relative path from.
        If not provided, "relative_path" is not included in the properties.

    :return: A dictionary with the extracted properties.
    Dictionary keys:
        - idx: The index of the data folder.
        - name: The name of the data folder.
        - datetime attributes "year", "month", "day", "hour", "minute", "second".
        - path: The absolute path of the data folder.
        - relative_path: The relative path of the data folder w.r.t the root_data_folder.
    """
    pattern = pattern.replace("{idx}", r"(?P<idx>\d+)")
    pattern = pattern.replace("{name}", r"(?P<name>\w+)")
    pattern = pattern.replace("%Y", r"(?P<year>\d{4})")
    pattern = pattern.replace("%m", r"(?P<month>\d{2})")
    pattern = pattern.replace("%d", r"(?P<day>\d{2})")
    pattern = pattern.replace("%H", r"(?P<hour>\d{2})")
    pattern = pattern.replace("%M", r"(?P<minute>\d{2})")
    pattern = pattern.replace("%S", r"(?P<second>\d{2})")

    data_folder = Path(data_folder)
    if root_data_folder is not None:
        folder_path_str = str(data_folder.relative_to(root_data_folder))
    else:
        folder_path_str = str(data_folder)

    folder_path_str = folder_path_str.replace("\\", "/")

    regex_match = re.match(pattern, folder_path_str)
    if regex_match is None:
        return None
    properties = regex_match.groupdict()
    properties = {key: int(value) if value.isdigit() else value for key, value in properties.items()}

    datetime_properties = {
        key: properties.pop(key) for key in ["year", "month", "day", "hour", "minute", "second"] if key in properties
    }
    properties["created_at"] = datetime(**datetime_properties)

    properties["path"] = data_folder
    if root_data_folder is not None:
        properties["relative_path"] = data_folder.relative_to(root_data_folder)
    return properties


def get_latest_data_folder(
    root_data_folder: Path,
    folder_pattern: str = DEFAULT_FOLDER_PATTERN,
    relative_path: Path = Path("."),
    current_folder_pattern: str = None,
) -> Optional[Dict[str, Union[str, int]]]:
    """Get the latest data folder in a given root data folder.

    Typically this is the folder within a date folder with the highest index.

    :param root_data_folder: The root data folder to search for the latest data folder.
    :param folder_pattern: The pattern of the data folder, e.g. "%Y-%m-%d/#{idx}_{name}_%H%M%S".
    :param relative_path: The relative path to the data folder. Used for recursive calls.
    :param current_folder_pattern: The current folder pattern. Used for recursive calls.
    :return: A dictionary with the properties of the latest data folder.
    Dictionary keys:
        - idx: The index of the data folder.
        - name: The name of the data folder.
        - datetime attributes "year", "month", "day", "hour", "minute", "second".
        - path: The absolute path of the data folder.
        - relative_path: The relative path of the data folder w.r.t the root_data_folder.
    """
    if isinstance(root_data_folder, str):
        root_data_folder = Path(root_data_folder)

    if not root_data_folder.exists():
        try:
            root_data_folder.mkdir(parents=True)
            print(f"The provided root folder didn't exist, so it was created at '{root_data_folder.absolute()}'.")
        except (Exception,):
            raise NotADirectoryError(f"Root data folder {root_data_folder} does not exist.")

    if current_folder_pattern is None:
        current_folder_pattern = folder_pattern

    current_folder_pattern, *remaining_folder_pattern = current_folder_pattern.split("/", maxsplit=1)

    folder_path = root_data_folder / relative_path

    if not remaining_folder_pattern:
        if "{idx}" not in current_folder_pattern:
            raise ValueError("The folder pattern must contain '{idx}' at the end.")

        folders = {}
        for f in folder_path.iterdir():
            if not f.is_dir():
                continue
            properties = extract_data_folder_properties(f, folder_pattern, root_data_folder=root_data_folder)
            if properties is None:
                continue

            folders[f] = properties
        if not folders:
            return None

        latest_folder, latest_properties = max(folders.items(), key=lambda f: f[1]["idx"])
        return latest_properties
    elif "{idx}" in current_folder_pattern:
        raise ValueError("The folder pattern must only contain '{idx}' in the last part.")
    else:
        # Filter out elements that aren't folders
        folders = filter(lambda f: f.is_dir(), folder_path.iterdir())
        # Filter folders that match the datetime of the current folder pattern
        folders = filter(lambda f: _validate_datetime(f.name, current_folder_pattern), folders)

        if not folders:
            return None

        # Sort folders by name (either datetime or index)
        sorted_folders = sorted(folders, key=lambda f: f.name, reverse=True)

        # Iterate over the folders, recursively call determine_latest_data_folder_idx
        for folder in sorted_folders:
            sub_folder_properties = get_latest_data_folder(
                root_data_folder,
                folder_pattern=folder_pattern,
                current_folder_pattern=remaining_folder_pattern[0],
                relative_path=relative_path / folder.name,
            )
            if sub_folder_properties is not None:
                return sub_folder_properties
        return None


def generate_data_folder_relative_pathname(
    idx: int, name: str, created_at: datetime, folder_pattern: str = DEFAULT_FOLDER_PATTERN
) -> str:
    """Generate the relative data folder name using a folder pattern.

    :param idx: The index of the data folder.
    :param name: The name of the data folder.
    :param created_at: The datetime to use for the folder name.
    :param folder_pattern: The pattern of the data folder, e.g. "%Y-%m-%d/#{idx}_{name}_%H%M%S".
    :return: The relative folder name
    """
    relative_folder_name = folder_pattern.format(idx=idx, name=name)
    relative_folder_name = created_at.strftime(relative_folder_name)
    return relative_folder_name


def create_data_folder(
    root_data_folder: Path,
    name: str,
    idx: Optional[int] = None,
    folder_pattern: str = DEFAULT_FOLDER_PATTERN,
    created_at: Optional[datetime] = None,
    create: bool = True,
) -> Dict[str, Union[str, int, Path]]:
    """Create a new data folder in a given root data folder.

    First checks the index of the latest data folder and increments by one.

    :param root_data_folder: The root data folder to create the new data folder in.
    :param name: The name of the new data folder.
    :param idx: The index of the new data folder. If not provided, the index is determined automatically.
    :param folder_pattern: The pattern of the data folder, e.g. "%Y-%m-%d/#{idx}_{name}_%H%M%S".
    :param created_at: The datetime to use for the folder name.
    :param create: Whether to create the folder or not.
    :return: A dictionary with the properties of the new data folder.
    Dictionary keys:
        - idx: The index of the data folder.
        - name: The name of the data folder.
        - path: The absolute path of the data folder.
        - relative_path: The relative path of the data folder w.r.t the root_data_folder.
        - created_at: The datetime the data folder was created.
    """
    if isinstance(root_data_folder, str):
        root_data_folder = Path(root_data_folder)

    if not root_data_folder.exists():
        raise NotADirectoryError(f"Root data folder {root_data_folder} does not exist.")

    if created_at is None:
        created_at = datetime.now()

    if idx is None:
        # Determine the latest folder index and increment by one
        latest_folder_properties = get_latest_data_folder(root_data_folder, folder_pattern=folder_pattern)

        if latest_folder_properties is None:
            # Create new folder with index 1
            idx = 1
        else:
            idx = latest_folder_properties["idx"] + 1

    relative_folder_name = generate_data_folder_relative_pathname(
        idx=idx, name=name, created_at=created_at, folder_pattern=folder_pattern
    )

    data_folder = root_data_folder / relative_folder_name

    if not create:
        return {
            "idx": idx,
            "name": name,
            "path": data_folder,
            "relative_path": data_folder.relative_to(root_data_folder),
            "created_at": created_at,
        }

    if data_folder.exists():
        raise FileExistsError(f"Data folder {data_folder} already exists.")

    data_folder.mkdir(parents=True)

    # TODO Remove datetime entries from properties
    properties = extract_data_folder_properties(data_folder, folder_pattern, root_data_folder)
    if properties is None:
        raise ValueError(f"Could not extract properties from data folder {data_folder}.")

    return properties



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from dash import dcc
import dash_bootstrap_components as dbc
from dash import html
from .app import app
import os
from .upload import UPLOAD_DIRECTORY

from dash import Input, Output

__all__ = ["download_page"]


def generate_download_page():
    return html.Div(
        [
            html.P("Please download both files"),
            html.Div(
                [
                    dbc.Button("config_initial.py", id="download-config-inital"),
                    dcc.Download(id="config-initial-data"),
                    html.Span(" + ", style={"fontSize": "3rem", "verticalAlign": "middle"}),
                    dbc.Button("config_edits.py", id="download-config-edits"),
                    dcc.Download(id="config-edits-data"),
                ]
            ),
            html.P("and use it as"),
            html.P(html.Code("from config_edits import configuration, setup")),
            html.P(html.Code("configuration = setup.build(configuration)")),
        ]
    )


download_page = generate_download_page()


@app.callback(
    Output("config-initial-data", "data"),
    Input("download-config-inital", "n_clicks"),
    prevent_initial_call=True,
)
def intial_config_download(n_clicks):
    return dcc.send_file(os.path.join(UPLOAD_DIRECTORY, "config_initial.py"))


@app.callback(
    Output("config-edits-data", "data"),
    Input("download-config-edits", "n_clicks"),
    prevent_initial_call=True,
)
def edit_config_download(n_clicks):
    return dcc.send_file(os.path.join(UPLOAD_DIRECTORY, "config_edits.py"))



from typing import Tuple, Union

import numpy as np


def _beta(v, u):
    lut = [[0, 0, 0, 0], [0, 0, 3, 1], [0, 1, 0, 3], [0, 3, 1, 0]]
    lut_map = {(0, 0): 0, (1, 0): 1, (0, 1): 2, (1, 1): 3}
    n = len(u) // 2
    beta = 0
    for i in range(n):
        beta += lut[lut_map[tuple(v[2 * i : 2 * i + 2])]][lut_map[tuple(u[2 * i : 2 * i + 2])]]
    return beta


def _compose_alpha(g1, alpha1, g2, alpha2):
    n = len(alpha1) // 2
    alpha21 = []
    two_alpha21 = np.zeros(2 * n, dtype=np.uint8)
    for i in range(2 * n):
        b_i = _calc_b_i(g1, g2, i)
        two_alpha21[i] += (2 * alpha1[i] + 2 * np.dot(g1[:, i], alpha2) + b_i) % 4
        assert two_alpha21[i] % 2 == 0
        alpha21.append(two_alpha21[i] // 2)
    return np.array(alpha21).astype(np.uint8)


def _calc_b_i(g1, g2, i):
    # add i for every Y in the column of g1
    b_i = np.dot(g1[::2, i], g1[1::2, i]) % 4
    n = g1.shape[0] // 2

    # reduce using beta
    current = np.zeros(2 * n, dtype=np.uint8)
    for j in range(2 * n):
        b_i = (b_i + _beta(current, g1[j, i] * g2[:, j])) % 4
        current = (current + g1[j, i] * g2[:, j]) % 2
    return b_i


def _calc_inverse_alpha(g1, alpha1):
    n = len(alpha1) // 2
    lam = _lambda(n)
    inv_g1 = lam @ g1.T @ lam % 2
    b = np.array([_calc_b_i(g1, inv_g1, i) for i in range(2 * n)])
    two_alpha2 = -(inv_g1.T @ (2 * alpha1 + b)) % 4
    assert np.all(two_alpha2 % 2 == 0)
    return two_alpha2 // 2


class SimpleTableau:
    """
    A class for representing and acting of Clifford Tableaus using only elementary logical and
    arithmetic operations. Implements the `CliffordTableau` API.
    """

    def __init__(self, g, alpha):
        g = np.array(g, dtype=np.uint8)
        alpha = np.array(alpha, dtype=np.uint8)
        if not len(alpha) % 2 == 0:
            raise ValueError(f"alpha needs to have an even length but length is {len(alpha)}")
        g_shape = g.shape
        if not (len(g_shape) == 2 and g_shape[0] == g_shape[1] and g_shape[0] % 2 == 0):
            raise ValueError(f"g has shape {g_shape}, which is not an even square matrix")
        if not len(alpha) % 2 == 0:
            raise ValueError(f"alpha has len {len(alpha)}, which is not even")
        self._n = len(alpha) // 2
        if not _is_symplectic(g, self._n):
            raise ValueError("g is not a symplectic matrix")
        self._np_repr = np.vstack((g, alpha)).astype(np.uint8)

    def __call__(self, *args, **kwargs):
        raise NotImplementedError()

    @property
    def g(self):
        return self._np_repr[:-1, :]

    @property
    def alpha(self):
        return self._np_repr[-1]

    @property
    def n(self):
        return self._n

    def __str__(self):
        # todo: work with more than single digit qubit number
        st = " " * 2 + "|" + " ".join(f"x{i} z{i}" for i in range(self._n)) + "\n"
        st += "-" * 2 + "+" + "-" * 6 * self._n + "\n"
        for i in range(2 * self._n):
            st += f"z{i // 2}|" if i % 2 else f"x{i // 2}|"
            st += "  ".join(str(entry) for entry in self._np_repr[i]) + "\n"
        st += "s |" + "  ".join(str(entry) for entry in self._np_repr[i + 1]) + "\n"
        return st

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return np.array_equal(self._np_repr, other._np_repr)

    def __hash__(self):
        return hash(self._np_repr.tobytes())

    def then(self, other: "SimpleTableau") -> "SimpleTableau":
        if self.n != other.n:
            raise ValueError(f"number of qubits of self={self.n} and of other={other.n} is incompatible")
        g12 = other.g @ self.g % 2
        alpha12 = _compose_alpha(self.g, self.alpha, other.g, other.alpha)
        return SimpleTableau(g12, alpha12)

    def inverse(self) -> "SimpleTableau":
        lam = _lambda(self.n)
        return SimpleTableau((lam @ self.g.T @ lam) % 2, _calc_inverse_alpha(self.g, self.alpha))

    def is_identity(self):
        return np.array_equal(self.g, np.eye(2 * self.n)) and np.array_equal(self.alpha, np.zeros(2 * self.n))


_single_qubit_gate_conversions = {
    "I": (np.identity(2), np.zeros(2)),
    "H": (np.array([[0, 1], [1, 0]]), np.zeros(2)),
    "X": (np.identity(2), np.array([0, 1])),
    "Z": (np.identity(2), np.array([1, 0])),
    "Y": (np.identity(2), np.array([1, 1])),
    "S": (np.array([[1, 0], [1, 1]]), np.zeros(2)),
    "SX": (np.array([[1, 1], [0, 1]]), np.array([0, 1])),
    "SY": (np.array([[0, 1], [1, 0]]), np.array([1, 0])),
    "-SY": (np.array([[0, 1], [1, 0]]), np.array([0, 1])),
    "-SX": (np.array([[1, 1], [0, 1]]), np.array([0, 0])),
}

_two_qubit_gate_conversions = {
    "CNOT": (np.array([[1, 0, 1, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 1, 0, 1]]).T, np.zeros(4)),
    "ISWAP": (
        np.array([[0, 1, 1, 1], [1, 1, 0, 1], [0, 0, 0, 1], [0, 0, 1, 0]]).T,
        np.zeros(4),
    ),  # not 100% sure this is correct
    "SWAP": (np.array([[0, 0, 1, 0], [0, 0, 0, 1], [1, 0, 0, 0], [0, 1, 0, 0]]).T, np.zeros(4)),
    "CZ": (np.array([[1, 0, 0, 1], [0, 1, 0, 0], [0, 1, 1, 0], [0, 0, 0, 1]]).T, np.zeros(4)),
}

known_names = set(_single_qubit_gate_conversions.keys()).union(set(_two_qubit_gate_conversions.keys()))


def generate_from_name(name: str, target: Union[int, Tuple[int, int]], n=None) -> SimpleTableau:
    """
    Generate a `SimpleTableau` object from a name.
    Args:
        name: gate name, must be in `known_names`
        target: For a single qubit gate, the qubit on which the gate operates, from 0 to n-1. For two qubit gates,
                a tuple of the form (control, target) on which the gate operates
        n: number of total qubits on which gate operates. If omitted, will be determined by the highest target index.

    Returns: the `SimpleTableau` object

    """
    if n is None:
        n = np.max(target) + 1
    g = np.identity(2 * n)
    alpha = np.zeros(2 * n)
    if name in _single_qubit_gate_conversions:
        if not isinstance(target, int) or target >= n:
            raise ValueError(f"invalid target {target} for single qubit gate {name} on {n} qubits")
        _embed_single_qubit_gate(alpha, g, name, target)
        return SimpleTableau(g, alpha)

    elif name in _two_qubit_gate_conversions:
        if not isinstance(target, tuple) or max(target) >= n:
            raise ValueError(f"invalid target {target} for two qubit gate {name} on {n} qubits")
        _embed_two_qubit_gate(alpha, g, name, target)
        return SimpleTableau(g, alpha)
    else:
        raise ValueError(f"unknown gate {name}")


def _embed_two_qubit_gate(alpha, g, name, target):
    g2q, alpha2q = _two_qubit_gate_conversions[name]
    g[2 * target[0] : 2 * target[0] + 2, 2 * target[0] : 2 * target[0] + 2] = g2q[:2, :2]
    g[2 * target[0] : 2 * target[0] + 2, 2 * target[1] : 2 * target[1] + 2] = g2q[:2, 2:4]
    g[2 * target[1] : 2 * target[1] + 2, 2 * target[0] : 2 * target[0] + 2] = g2q[2:4, :2]
    g[2 * target[1] : 2 * target[1] + 2, 2 * target[1] : 2 * target[1] + 2] = g2q[2:4, 2:4]
    alpha[2 * target[0] : 2 * target[0] + 2] = alpha2q[:2]
    alpha[2 * target[1] : 2 * target[1] + 2] = alpha2q[2:4]


def _embed_single_qubit_gate(alpha, g, name, target):
    g[2 * target : 2 * target + 2, 2 * target : 2 * target + 2] = _single_qubit_gate_conversions[name][0]
    alpha[2 * target : 2 * target + 2] = _single_qubit_gate_conversions[name][1]


def _lambda(n):
    return np.diag([1] + [0, 1] * (n - 1), 1) + np.diag([1] + [0, 1] * (n - 1), -1)


def _is_symplectic(mat, n):
    lhs = mat @ _lambda(n) @ mat.T % 2
    return np.all(lhs == _lambda(n))


_int_bit_map = {0: [0, 0], 1: [1, 0], 2: [1, 1], 3: [0, 1]}


def stim_to_simple(tableau) -> SimpleTableau:
    n = len(tableau)
    g = np.zeros((2 * n, 2 * n), dtype=np.uint8)
    alpha = np.zeros(2 * n, dtype=np.uint8)
    for j in range(0, 2 * n, 2):
        pauli_string_x = tableau.x_output(j // 2)
        pauli_string_z = tableau.z_output(j // 2)
        alpha[j] = pauli_string_x.sign.real < 0
        alpha[j + 1] = pauli_string_z.sign.real < 0
        for i in range(0, 2 * n, 2):
            g[i : i + 2, j] = _int_bit_map[pauli_string_x[i // 2]]
            g[i : i + 2, j + 1] = _int_bit_map[pauli_string_z[i // 2]]
    return SimpleTableau(g, alpha)



import warnings
import re


class ConfigurationError(Exception):
    def __init__(self, message) -> None:
        self.message = message
        super().__init__(self.message)


def _warn_if_not_none(kwargs, deprecated_arg):
    suggested_arg = "element_" + re.split("_ports", deprecated_arg)[0] + "s"
    val = kwargs.get(deprecated_arg, None)
    if val is not None:
        warnings.warn(
            "{0} is deprecated use {1}".format(deprecated_arg, suggested_arg),
            DeprecationWarning,
        )
    return val



def import_cirq():
    try:
        import cirq

        return cirq
    except ImportError:
        raise ImportError(
            "Cirq is not installed. Please install it with:\n"
            "pip install cirq==1.3.0\n"
            "Note: Cirq requires Python >= 3.10. Please ensure your Python version meets this requirement."
        )



import numpy as np
from scipy.signal.windows import gaussian, blackman


def drag_gaussian_pulse_waveforms(
    amplitude, length, sigma, alpha, anharmonicity, detuning=0.0, subtracted=True, sampling_rate=1e9, **kwargs
):
    """
    Creates Gaussian based DRAG waveforms that compensate for the leakage and for the AC stark shift.

    These DRAG waveforms has been implemented following the next Refs.:
    Chen et al. PRL, 116, 020501 (2016)
    https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.116.020501
    and Chen's thesis
    https://web.physics.ucsb.edu/~martinisgroup/theses/Chen2018.pdf

    :param float amplitude: The amplitude in volts.
    :param int length: The pulse length in ns.
    :param float sigma: The gaussian standard deviation.
    :param float alpha: The DRAG coefficient.
    :param float anharmonicity: f_21 - f_10 - The differences in energy between the 2-1 and the 1-0 energy levels, in Hz.
    :param float detuning: The frequency shift to correct for AC stark shift, in Hz.
    :param bool subtracted: If true, returns a subtracted Gaussian, such that the first and last points will be at 0
        volts. This reduces high-frequency components due to the initial and final points offset. Default is true.
    :param float sampling_rate: The sampling rate used to describe the waveform, in samples/s. Default is 1G samples/s.
    :return: Returns a tuple of two lists. The first list is the 'I' waveform (real part) and the second is the
        'Q' waveform (imaginary part)
    """
    if alpha != 0 and anharmonicity == 0:
        raise Exception("Cannot create a DRAG pulse with `anharmonicity=0`")
    t = np.arange(length, step=1e9 / sampling_rate)  # An array of size pulse length in ns
    center = (length - 1e9 / sampling_rate) / 2
    gauss_wave = amplitude * np.exp(-((t - center) ** 2) / (2 * sigma**2))  # The gaussian function
    gauss_der_wave = (
        amplitude * (-2 * 1e9 * (t - center) / (2 * sigma**2)) * np.exp(-((t - center) ** 2) / (2 * sigma**2))
    )  # The derivative of gaussian
    if subtracted:
        gauss_wave = gauss_wave - gauss_wave[-1]  # subtracted gaussian
    z = gauss_wave + 1j * 0
    if anharmonicity != detuning:
        # The complex DRAG envelope:
        z += 1j * gauss_der_wave * (alpha / (2 * np.pi * anharmonicity - 2 * np.pi * detuning))
    elif alpha != 0:
        raise ValueError(
            "The complex envelop for the DRAG waveform cannot be created if anharmonicity = detuning and alpha != 0."
        )
    # The complex detuned DRAG envelope:
    z *= np.exp(1j * 2 * np.pi * detuning * t * 1e-9)
    I_wf = z.real.tolist()  # The `I` component is the real part of the waveform
    Q_wf = z.imag.tolist()  # The `Q` component is the imaginary part of the waveform
    return I_wf, Q_wf


def drag_cosine_pulse_waveforms(amplitude, length, alpha, anharmonicity, detuning=0.0, sampling_rate=1e9, **kwargs):
    """
    Creates Cosine based DRAG waveforms that compensate for the leakage and for the AC stark shift.

    These DRAG waveforms has been implemented following the next Refs.:
    Chen et al. PRL, 116, 020501 (2016)
    https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.116.020501
    and Chen's thesis
    https://web.physics.ucsb.edu/~martinisgroup/theses/Chen2018.pdf

    :param float amplitude: The amplitude in volts.
    :param int length: The pulse length in ns.
    :param float alpha: The DRAG coefficient.
    :param float anharmonicity: f_21 - f_10 - The differences in energy between the 2-1 and the 1-0 energy levels, in Hz.
    :param float detuning: The frequency shift to correct for AC stark shift, in Hz.
    :param float sampling_rate: The sampling rate used to describe the waveform, in samples/s. Default is 1G samples/s.
    :return: Returns a tuple of two lists. The first list is the 'I' waveform (real part) and the second is the
        'Q' waveform (imaginary part)
    """
    if alpha != 0 and anharmonicity == 0:
        raise Exception("Cannot create a DRAG pulse with `anharmonicity=0`")
    t = np.arange(length, step=1e9 / sampling_rate)  # An array of size pulse length in ns
    end_point = length - 1e9 / sampling_rate
    cos_wave = 0.5 * amplitude * (1 - np.cos(t * 2 * np.pi / end_point))  # The cosine function
    sin_wave = (
        0.5 * amplitude * (2 * np.pi / end_point * 1e9) * np.sin(t * 2 * np.pi / end_point)
    )  # The derivative of cosine function
    z = cos_wave + 1j * 0
    if anharmonicity != detuning:
        # The complex DRAG envelope:
        z += 1j * sin_wave * (alpha / (2 * np.pi * anharmonicity - 2 * np.pi * detuning))
    elif alpha != 0:
        raise ValueError(
            "The complex envelop for the DRAG waveform cannot be created if anharmonicity = detuning and alpha != 0."
        )
    # The complex detuned DRAG envelope:
    z *= np.exp(1j * 2 * np.pi * detuning * t * 1e-9)
    I_wf = z.real.tolist()  # The `I` component is the real part of the waveform
    Q_wf = z.imag.tolist()  # The `Q` component is the imaginary part of the waveform
    return I_wf, Q_wf


def flattop_gaussian_waveform(amplitude, flat_length, rise_fall_length, return_part="all", sampling_rate=1e9):
    """
    Returns a flat top Gaussian waveform. This is a square pulse with a rise and fall of a Gaussian with the given
    sigma. It is possible to only get the rising or falling parts, which allows scanning the flat part length from QUA.
    The length of the pulse will be the `flat_length + 2 * rise_fall_length`.

    :param float amplitude: The amplitude in volts.
    :param int flat_length: The flat part length in ns.
    :param int rise_fall_length: The rise and fall times in ns. The Gaussian sigma is given by the
        `rise_fall_length / 5`.
    :param str return_part: When set to 'all', returns the complete waveform. Default is 'all'. When set to 'rise',
    returns only the rising part. When set to 'fall', returns only the falling part. This is useful for separating
    the three parts which allows scanning the duration of the flat part is to scanned from QUA
    :param float sampling_rate: The sampling rate used to describe the waveform, in samples/s. Must be an integer multiple of 1e9 samples per seconds. Default is 1G samples/s.
    :return: Returns the waveform as a list of values with 1ns spacing
    """

    assert sampling_rate % 1e9 == 0, "The sampling rate must be an integer multiple of 1e9 samples per second."

    gauss_wave = amplitude * gaussian(
        int(np.round(2 * rise_fall_length * sampling_rate / 1e9)), rise_fall_length / 5 * sampling_rate / 1e9
    )
    rise_part = gauss_wave[: int(rise_fall_length * sampling_rate / 1e9)]
    rise_part = rise_part.tolist()
    if return_part == "all":
        return rise_part + [amplitude] * int(flat_length * sampling_rate / 1e9) + rise_part[::-1]
    elif return_part == "rise":
        return rise_part
    elif return_part == "fall":
        return rise_part[::-1]
    else:
        raise Exception("'return_part' must be either 'all', 'rise' or 'fall'")


def flattop_cosine_waveform(amplitude, flat_length, rise_fall_length, return_part="all", sampling_rate=1e9):
    """
    Returns a flat top cosine waveform. This is a square pulse with a rise and fall with cosine shape with the given
    sigma. It is possible to only get the rising or falling parts, which allows scanning the flat part length from QUA.
    The length of the pulse will be the `flat_length + 2 * rise_fall_length`.

    :param float amplitude: The amplitude in volts.
    :param int flat_length: The flat part length in ns.
    :param int rise_fall_length: The rise and fall times in ns, taken as the time for a cosine to go from 0 to 1
    (pi phase-shift) and conversely.
    :param str return_part: When set to 'all', returns the complete waveform. Default is 'all'. When set to 'rise',
    returns only the rising part. When set to 'fall', returns only the falling part. This is useful for separating
    the three parts which allows scanning the duration of the flat part is to scanned from QUA
    :param float sampling_rate: The sampling rate used to describe the waveform, in samples/s. Must be an integer multiple of 1e9 samples per seconds. Default is 1G samples/s.
    :return: Returns the waveform as a list of values with 1ns spacing
    """
    assert sampling_rate % 1e9 == 0, "The sampling rate must be an integer multiple of 1e9 samples per second."
    rise_part = amplitude * 0.5 * (1 - np.cos(np.linspace(0, np.pi, int(rise_fall_length * sampling_rate / 1e9))))
    rise_part = rise_part.tolist()
    if return_part == "all":
        return rise_part + [amplitude] * int(flat_length * sampling_rate / 1e9) + rise_part[::-1]
    elif return_part == "rise":
        return rise_part
    elif return_part == "fall":
        return rise_part[::-1]
    else:
        raise Exception("'return_part' must be either 'all', 'rise' or 'fall'")


def flattop_tanh_waveform(amplitude, flat_length, rise_fall_length, return_part="all", sampling_rate=1e9):
    """
    Returns a flat top tanh waveform. This is a square pulse with a rise and fall with tanh shape with the given
    sigma. It is possible to only get the rising or falling parts, which allows scanning the flat part length from QUA.
    The length of the pulse will be the `flat_length + 2 * rise_fall_length`.

    :param float amplitude: The amplitude in volts.
    :param int flat_length: The flat part length in ns.
    :param int rise_fall_length: The rise and fall times in ns, taken as a number of points of a tanh between -4
        and 4.
    :param str return_part: When set to 'all', returns the complete waveform. Default is 'all'. When set to 'rise',
    returns only the rising part. When set to 'fall', returns only the falling part. This is useful for separating
    the three parts which allows scanning the duration of the flat part is to scanned from QUA
    :param float sampling_rate: The sampling rate used to describe the waveform, in samples/s. Must be an integer multiple of 1e9 samples per seconds. Default is 1G samples/s.
    :return: Returns the waveform as a list of values with 1ns spacing
    """
    assert sampling_rate % 1e9 == 0, "The sampling rate must be an integer multiple of 1e9 samples per second."
    rise_part = amplitude * 0.5 * (1 + np.tanh(np.linspace(-4, 4, int(rise_fall_length * sampling_rate / 1e9))))
    rise_part = rise_part.tolist()
    if return_part == "all":
        return rise_part + [amplitude] * int(flat_length * sampling_rate / 1e9) + rise_part[::-1]
    elif return_part == "rise":
        return rise_part
    elif return_part == "fall":
        return rise_part[::-1]
    else:
        raise Exception("'return_part' must be either 'all', 'rise' or 'fall'")


def flattop_blackman_waveform(amplitude, flat_length, rise_fall_length, return_part="all", sampling_rate=1e9):
    """
    Returns a flat top Blackman waveform. This is a square pulse with a rise and fall with Blackman shape with the given
    length. It is possible to only get the rising or falling parts, which allows scanning the flat part length from QUA.
    The length of the pulse will be the `flat_length + 2 * rise_fall_length`.

    :param float amplitude: The amplitude in volts.
    :param int flat_length: The flat part length in ns.
    :param int rise_fall_length: The rise and fall times in ns, taken as the time to go from 0 to 'amplitude'.
    :param str return_part: When set to 'all', returns the complete waveform. Default is 'all'. When set to 'rise',
    returns only the rising part. When set to 'fall', returns only the falling part. This is useful for separating
    the three parts which allows scanning the duration of the  flat part is to scanned from QUA
    :param float sampling_rate: The sampling rate used to describe the waveform, in samples/s. Must be an integer multiple of 1e9 samples per seconds. Default is 1G samples/s.
    :return: Returns the waveform as a list
    """
    assert sampling_rate % 1e9 == 0, "The sampling rate must be an integer multiple of 1e9 samples per second."
    backman_wave = amplitude * blackman(2 * int(rise_fall_length * sampling_rate / 1e9))
    rise_part = backman_wave[: int(rise_fall_length * sampling_rate / 1e9)]
    rise_part = rise_part.tolist()
    if return_part == "all":
        return rise_part + [amplitude] * int(flat_length * sampling_rate / 1e9) + rise_part[::-1]
    elif return_part == "rise":
        return rise_part
    elif return_part == "fall":
        return rise_part[::-1]
    else:
        raise Exception("'return_part' must be either 'all', 'rise' or 'fall'")


def blackman_integral_waveform(pulse_length, v_start, v_end, sampling_rate=1e9):
    """
    Returns a Blackman integral waveform. This is the integral of a Blackman waveform, adiabatically going from
    'v_start' to 'v_end' in 'pulse_length' ns.

    :param int pulse_length: The pulse length in ns.
    :param float v_start: The starting amplitude in volts.
    :param float v_end: The ending amplitude in volts.
    :param float sampling_rate: The sampling rate used to describe the waveform, in samples/s. Must be an integer multiple of 1e9 samples per seconds. Default is 1G samples/s.
    :return: Returns the waveform as a list
    """
    assert sampling_rate % 1e9 == 0, "The sampling rate must be an integer multiple of 1e9 samples per second."
    time = np.linspace(0, pulse_length - 1, int(pulse_length * sampling_rate / 1e9))
    black_wave = v_start + (
        time / (pulse_length - 1)
        - (25 / (42 * np.pi)) * np.sin(2 * np.pi * time / (pulse_length - 1))
        + (1 / (21 * np.pi)) * np.sin(4 * np.pi * time / (pulse_length - 1))
    ) * (v_end - v_start)
    return black_wave.tolist()



from typing import List

import numpy as np
from matplotlib import pyplot as plt
from qm import Program, QuantumMachinesManager
from qm.qua import *
from tqdm import tqdm

from .util import run_in_thread
from .TwoQubitRB import TwoQubitRb
from .verification import SequenceTracker


phased_xz_command_sequences = {
    r"I \otimes I": [720],  # Identity on both qubits
    r"I \otimes Z": [732],  # Z on qubit 1, Identity on qubit 2
    r"Z \otimes I": [723],  # Z on qubit 2, Identity on qubit 1
    r"I \otimes X": [724],  # X on qubit 1, Identity on qubit 2
    r"X \otimes I": [721],  # X on qubit 2, Identity on qubit 1
    r"X \otimes X": [725],  # X on both qubits
    r"\frac{X}{2} \otimes I": [1],  # X/2 on qubit 2, Identity on qubit 1
    r"I \otimes \frac{X}{2}": [6],  # X/2 on qubit 1, Identity on qubit 2
    r"\frac{X}{2} \otimes \frac{X}{2}": [7],  # X/2 on both qubits
    r"\text{CZ}": [74],  # Controlled-Z (CZ) gate
    r"(\frac{Y}{2} \otimes -\frac{Y}{2}), \text {CZ}, (I \otimes \frac{Y}{2}) \Rightarrow |\Phi^+\rangle_{Bell}": [252],
    r"\text{CNOT}": [12, 347],  # X/2 on qubit 2, followed by CNOT
    r"(\frac{X}{2} \otimes I), \text{CNOT}": [1, 4, 63],  # X/2 on qubit 2, followed by CNOT
    r"(X \otimes I), \text{CNOT}": [724, 4, 63],  # X/2 on qubit 2, followed by CNOT
    r"(I \otimes X), \text{SWAP}": [724, 48, 498],  # X on qubit 1, followed by SWAP
}


class TwoQubitRbDebugger:
    def __init__(self, rb: TwoQubitRb):
        """
        Class which mimics the preparation, measurement and input streaming techniques
        used in TwoQubitRB to generate a program on a limited subset of gates and
        compares the results to expectation for debugging.
        """
        self.rb = rb

    def run_phased_xz_commands(self, qmm: QuantumMachinesManager, num_averages: int, unsafe: bool = False):
        """
        Run a program testing selected commands containing only combinations of PhasedXZ
        gates and other fundamental gates, which lead to a variety of transformations on
        the |00> state. This is useful for testing the 1Q component of the gate implementation.

        Params:
            unsafe (bool): Refers to the option of compiling a QUA switch-case "safely", which
                           guarantees correct behaviour but can lead to gaps, or "unsafely",
                           which reduces gaps but can cause unwanted behaviour. Note: as of
                           QOP 3.2.3, there is an issue with "unsafe" compilation in certain circumstances.

        """
        sequences = phased_xz_command_sequences.values()

        self.sequence_tracker = SequenceTracker(self.rb._command_registry)

        prog = self._phased_xz_commands_program(len(sequences), num_averages, unsafe)

        qm = qmm.open_qm(self.rb._config)
        job = qm.execute(prog)

        self._insert_all_input_stream(job, sequences)

        job.result_handles.wait_for_all_values()
        state = job.result_handles.get("state").fetch_all()

        self.sequence_tracker.print_sequences()
        self._analyze_phased_xz_commands_program(state, list(phased_xz_command_sequences.keys()))

    @run_in_thread
    def _insert_all_input_stream(self, job, sequences):
        for sequence in tqdm(sequences, desc="Running test-sequences", unit="sequence"):
            self.sequence_tracker.make_sequence(sequence)
            job.insert_input_stream("__gates_len_is__", len(sequence))
            for qe in self.rb._rb_baker.all_elements:
                job.insert_input_stream(
                    f"{self._input_stream_name_from_element(str(qe))}_is",
                    self.rb._decode_sequence_for_element(qe, sequence),
                )

    def _phased_xz_commands_program(self, num_sequences: int, num_averages: int, unsafe: bool) -> Program:
        with program() as prog:
            n_avg = declare(int)
            state = declare(int)
            length = declare(int)
            i = declare(int)
            state_os = declare_stream()
            gates_len_is = declare_input_stream(int, name="__gates_len_is__", size=1)
            gates_is = {
                qe: declare_input_stream(
                    int, name=f"{self._input_stream_name_from_element(str(qe))}_is", size=self.rb._buffer_length
                )
                for qe in self.rb._rb_baker.all_elements
            }

            with for_(i, 0, i < num_sequences, i + 1):
                advance_input_stream(gates_len_is)
                for gate_is in gates_is.values():
                    advance_input_stream(gate_is)
                assign(length, gates_len_is[0])
                with for_(n_avg, 0, n_avg < num_averages, n_avg + 1):
                    self.rb._prep_func()
                    self.rb._rb_baker.run(gates_is, length, unsafe=unsafe)
                    out1, out2 = self.rb._measure_func()
                    assign(state, (Cast.to_int(out2) << 1) + Cast.to_int(out1))
                    save(state, state_os)

            with stream_processing():
                state_os.buffer(num_averages).buffer(num_sequences).save("state")

        return prog

    def _analyze_phased_xz_commands_program(self, state: np.ndarray, sequence_labels: List[str]):
        fig, axs = plt.subplots(5, 3, figsize=(12, 10))
        axs = axs.ravel()

        basis_states = [r"$|00\rangle$", r"$|01\rangle$", r"$|10\rangle$", r"$|11\rangle$"]

        for i, sequence in enumerate(self.sequence_tracker._sequences_as_gates):
            expected_state = self.sequence_tracker.calculate_resultant_state(sequence)
            expected_distribution_into_two_qubit_bases = np.diag(expected_state)

            counts = np.bincount(state[i], minlength=4)
            counts = counts / counts.sum()

            axs[i].set_title(f"${sequence_labels[i]}$")
            axs[i].bar(basis_states, expected_distribution_into_two_qubit_bases, label="Expected", alpha=0.5, color="b")
            axs[i].bar(basis_states, counts, label="Measured", alpha=0.5, color="r")

        handles, labels = axs[-1].get_legend_handles_labels()
        fig.legend(handles, labels, loc="upper left", ncol=2)
        fig.suptitle(r"State Distribution of PhasedXZ Commands applied to $|00\rangle$")
        fig.supylabel("Measurement Probability")
        fig.supxlabel(r"Basis State as $|q_2, q_1\rangle$")

        plt.tight_layout()
        plt.plot()
        plt.show()

    def _input_stream_name_from_element(self, element_name: str):
        return element_name.replace(".", "_")



from typing import List, Type

from .data_processor import DataProcessor
from .matplotlib_plot_saver import MatplotlibPlotSaver
from .numpy_array_saver import NumpyArraySaver
from .xarray_saver import XarraySaver
from .waveform_report_saver import WaveformReportSaver
from .simulator_controller_samples_saver import SimulatorControllerSamplesSaver
from .plotly_graph_saver import PlotlyGraphSaver

DEFAULT_DATA_PROCESSORS: List[Type[DataProcessor]] = [
    DataProcessor,
    WaveformReportSaver,
    SimulatorControllerSamplesSaver,
    MatplotlibPlotSaver,
    NumpyArraySaver,
]

try:
    import xarray  # noqa: F401

    DEFAULT_DATA_PROCESSORS.append(XarraySaver)
except ImportError:
    pass

try:
    import plotly  # noqa: F401

    DEFAULT_DATA_PROCESSORS.append(PlotlyGraphSaver)
except ImportError:
    pass


__all__ = [
    "DataProcessor",
    "MatplotlibPlotSaver",
    "NumpyArraySaver",
    "XarraySaver",
    "WaveformReportSaver",
    "SimulatorControllerSamplesSaver",
    "DEFAULT_DATA_PROCESSORS",
]



from __future__ import annotations

from qualang_tools.wirer.connectivity.channel_spec import ChannelSpec
from qualang_tools.wirer.connectivity.types import QubitsType, QubitPairsType, ElementsType
from qualang_tools.wirer.connectivity.wiring_spec import WiringFrequency, WiringIOType, WiringLineType
from qualang_tools.wirer.connectivity.connectivity_base import ConnectivityBase


class ConnectivityQuantumDotQubits(ConnectivityBase):
    """
    Represents the high-level wiring configuration for a quantum-dot-based QPU setup.

    This class defines and stores placeholders for quantum elements (e.g., quantum dots, sensor dots, and resonators)
    and specifies the wiring requirements for each of their control and readout lines. It enables
    the configuration of line types (e.g., drive, plunger gates, barrier gates, sensor gates), their I/O roles,
    and associated frequency domains (RF or DC), as well as constraints for channel allocation.

    The API supports a two-stage workflow:
    - Stage 1: define dot-layer wiring (gates and resonators) without qubit drive lines.
    - Stage 2: add qubit drive lines after building the qubit-layer QUAM.
    """

    def add_voltage_gate_lines(
        self,
        voltage_gates: ElementsType,
        triggered: bool = False,
        constraints: ChannelSpec | None = None,
        name: str = "vg",
        wiring_line_type: WiringLineType = WiringLineType.GLOBAL_GATE,
    ) -> None:
        """
        Adds specifications (placeholders) for generic voltage gate lines.

        This is a generic method for adding voltage gate lines with a specified line type. It can be used
        to add any type of voltage gate (global gates, sensor gates, etc.) by specifying the `wiring_line_type`.
        For more specific use cases, consider using the dedicated methods:
        - `add_quantum_dot_voltage_gate_lines`: For plunger gates on quantum dots
        - `add_quantum_dot_pair_voltage_gate_lines`: For barrier gates on quantum dot pairs
        - `add_sensor_dot_voltage_gate_lines`: For sensor gates on sensor dots

        No channels are allocated at this stage.

        Args:
            voltage_gates (ElementsType): The voltage gate elements to configure. Can be a single element ID
                or a list of element IDs.
            triggered (bool, optional): Whether the line is triggered. Defaults to False.
            constraints (ChannelSpec, optional): Constraints on the channel, if any. Defaults to None.
            name (str, optional): Name prefix for the elements. Defaults to 'vg'.
            wiring_line_type (WiringLineType, optional): The type of wiring line (e.g., GLOBAL_GATE, SENSOR_GATE).
                Defaults to GLOBAL_GATE.
        """
        elements = self._make_named_elements(name, voltage_gates)
        self.add_wiring_spec(
            WiringFrequency.DC, WiringIOType.OUTPUT, wiring_line_type, triggered, constraints, elements
        )

    def add_sensor_dots(
        self,
        sensor_dots: ElementsType,
        triggered: bool = False,
        constraints: ChannelSpec | None = None,
        shared_resonator_line: bool = False,
        use_mw_fem: bool = False,
    ) -> None:
        """
        Adds specifications (placeholders) for sensor dots, including both voltage gate lines and resonator lines.

        This method is a convenience function that adds both the voltage gate lines and resonator lines for sensor dots.
        It calls `add_sensor_dot_voltage_gate_lines` and `add_sensor_dot_resonator_line` internally.

        No channels are allocated at this stage.

        Args:
            sensor_dots (ElementsType): The sensor dot elements to configure. Can be a single element ID
                or a list of element IDs.
            triggered (bool, optional): Whether the lines are triggered. Defaults to False.
            constraints (ChannelSpec, optional): Constraints on the channels, if any. Defaults to None.
            shared_resonator_line (bool, optional): Whether multiple sensor dots share the same resonator line.
                Defaults to False.
            use_mw_fem (bool, optional): Whether to use MW-FEM for the resonator line. Defaults to False.
        """
        self.add_sensor_dot_voltage_gate_lines(sensor_dots, triggered=triggered, constraints=constraints)
        self.add_sensor_dot_resonator_line(
            sensor_dots,
            triggered=triggered,
            constraints=constraints,
            shared_line=shared_resonator_line,
            use_mw_fem=use_mw_fem,
        )

    def add_sensor_dot_voltage_gate_lines(
        self, sensor_dots: ElementsType, triggered: bool = False, constraints: ChannelSpec | None = None
    ):
        """
        Adds specifications (placeholders) for voltage gate lines on sensor dots.

        This method configures sensor gate line specifications (placeholders) for sensor dots, which are
        used for charge sensing in quantum dot systems. The gates are configured as SENSOR_GATE line type.

        No channels are allocated at this stage.

        Args:
            sensor_dots (ElementsType): The sensor dot elements to configure. Can be a single element ID
                or a list of element IDs.
            triggered (bool, optional): Whether the line is triggered. Defaults to False.
            constraints (ChannelSpec, optional): Constraints on the channel, if any. Defaults to None.
        """
        self.add_voltage_gate_lines(
            sensor_dots,
            triggered=triggered,
            constraints=constraints,
            name="s",
            wiring_line_type=WiringLineType.SENSOR_GATE,
        )

    def add_quantum_dots(
        self,
        quantum_dots: QubitsType,
        add_drive_lines: bool = False,
        triggered: bool = False,
        constraints: ChannelSpec | None = None,
        use_mw_fem=False,
        shared_drive_line=False,
    ):
        """
        Adds specifications (placeholders) for quantum dots

        This convenience method only adds plunger gate lines.

        No channels are allocated at this stage.

        Args:
            quantum_dots (QubitsType): The quantum dots to configure. Can be a single quantum dot ID or a list of quantum dot IDs.
            add_drive_lines (bool, optional): Whether to add drive lines. Defaults to False.
            triggered (bool, optional): Whether the line is triggered. Defaults to False.
            constraints (ChannelSpec, optional): Constraints on the channel, if any. Defaults to None.
            use_mw_fem (bool, optional): Whether to use MW-FEM for the drive line. Defaults to False.
            shared_drive_line (bool, optional): Whether multiple quantum dots share the same drive line.
                Defaults to False.
        """
        self.add_quantum_dot_voltage_gate_lines(quantum_dots)
        if add_drive_lines:

            self.add_quantum_dot_drive_lines(
                quantum_dots,
                triggered=triggered,
                constraints=constraints,
                use_mw_fem=use_mw_fem,
                shared_line=shared_drive_line,
            )

    def add_quantum_dot_pairs(
        self, quantum_dot_pairs: QubitPairsType, triggered: bool = False, constraints: ChannelSpec | None = None
    ):
        """
        Adds specifications (placeholders) for quantum dot pairs, including barrier gate lines.

        This method configures barrier gate line specifications (placeholders) for pairs of quantum dots, which are
        used to control the barrier between quantum dots in a pair. It calls `add_quantum_dot_pair_voltage_gate_lines`
        internally.

        No channels are allocated at this stage.

        Args:
            quantum_dot_pairs (QubitPairsType): The quantum dot pairs to configure. Can be a single pair (tuple of two quantum dot IDs)
                or a list of pairs.
            triggered (bool, optional): Whether the line is triggered. Defaults to False.
            constraints (ChannelSpec, optional): Constraints on the channel, if any. Defaults to None.
        """
        self.add_barrier_voltage_gate_lines(quantum_dot_pairs, triggered=triggered, constraints=constraints)

    def add_sensor_dot_resonator_line(
        self,
        sensor_dots,
        triggered: bool = False,
        constraints: ChannelSpec | None = None,
        shared_line: bool = False,
        use_mw_fem=False,
    ):
        """
        Adds a specification (placeholder) for a resonator line for sensor dots.

        This method configures a resonator line specification (placeholder) for sensor dots, typically used
        for readout via tank circuits. The default frequency is DC (LF-FEM) since most tank circuits operate
        below 800MHz and connect to LF-FEM. Only resonators above 2GHz can connect to MW-FEM due to bandwidth
        limitations. If you need to connect to MW-FEM, set `use_mw_fem=True`.

        No channels are allocated at this stage.

        Args:
            sensor_dots: The sensor dot elements to associate with the resonator line. Can be a single element ID
                or a list of element IDs.
            triggered (bool, optional): Whether the line is triggered. Defaults to False.
            constraints (ChannelSpec, optional): Constraints on the channel, if any. Defaults to None.
            shared_line (bool, optional): Whether multiple sensor dots share the same resonator line.
                Defaults to False.
            use_mw_fem (bool, optional): Whether to use MW-FEM for the resonator line. Defaults to False.

        Returns:
            List[WiringSpec]: A list of wiring specifications (placeholders) for the resonator lines.
        """
        elements = self._make_named_elements("s", sensor_dots)

        if use_mw_fem:
            wiring_frequency = WiringFrequency.RF
        else:
            wiring_frequency = WiringFrequency.DC

        return self.add_wiring_spec(
            wiring_frequency,
            WiringIOType.INPUT_AND_OUTPUT,
            WiringLineType.RF_RESONATOR,
            triggered,
            constraints,
            elements,
            shared_line=shared_line,
        )

    def add_quantum_dot_drive_lines(
        self,
        quantum_dots: QubitsType,
        triggered: bool = False,
        constraints: ChannelSpec | None = None,
        use_mw_fem=False,
        shared_line=False,
    ):
        """
        Adds specifications (placeholders) for RF drive lines for quantum dots.

        This method configures the drive line specifications (placeholders) for quantum dots,
        which are typically used to apply RF control signals for quantum dot manipulation. It allows optional
        triggering and constraints on which channel configurations can be allocated for this line.

        No channels are allocated at this stage.

        Args:
            quantum_dots (QubitsType): The quantum dots to configure the drive lines for. Can be a single quantum dot ID
                or a list of quantum dot IDs.
            triggered (bool, optional): Whether the line is triggered. Defaults to False.
            constraints (ChannelSpec, optional): Constraints on the channel, if any. Defaults to None.
            use_mw_fem (bool, optional): Whether to use MW-FEM for the drive line. Defaults to False.
            shared_line (bool, optional): Whether multiple quantum dots share the same drive line.
                Defaults to False.

        Returns:
            List[WiringSpec]: A list of wiring specifications (placeholders) for the quantum dot drive lines.
        """
        elements = self._make_qubit_elements(quantum_dots)
        if use_mw_fem:
            wiring_frequency = WiringFrequency.RF
        else:
            wiring_frequency = WiringFrequency.DC
        return self.add_wiring_spec(
            wiring_frequency,
            WiringIOType.OUTPUT,
            WiringLineType.DRIVE,
            triggered,
            constraints,
            elements,
            shared_line=shared_line,
        )

    def add_quantum_dot_voltage_gate_lines(
        self, quantum_dots: QubitsType, triggered: bool = False, constraints: ChannelSpec | None = None
    ):
        """
        Adds specifications (placeholders) for plunger gate lines on quantum dots.

        This method configures plunger gate line specifications (placeholders) for quantum dots.
        Plunger gates are used to control the charge occupation of quantum dots, allowing manipulation
        of the quantum dot state. The gates are configured as PLUNGER_GATE line type with DC frequency.

        This is distinct from `add_voltage_gate_lines` which is a generic method. Use this method when
        you specifically need plunger gates for quantum dots.

        No channels are allocated at this stage.

        Args:
            quantum_dots (QubitsType): The quantum dots to configure the plunger gate lines for. Can be a single quantum dot ID
                or a list of quantum dot IDs.
            triggered (bool, optional): Whether the line is triggered. Defaults to False.
            constraints (ChannelSpec, optional): Constraints on the channel, if any. Defaults to None.

        Returns:
            List[WiringSpec]: A list of wiring specifications (placeholders) for the plunger gate lines.
        """
        elements = self._make_qubit_elements(quantum_dots)
        return self.add_wiring_spec(
            WiringFrequency.DC, WiringIOType.OUTPUT, WiringLineType.PLUNGER_GATE, triggered, constraints, elements
        )

    def add_barrier_voltage_gate_lines(
        self, quantum_dot_pairs: QubitPairsType, triggered: bool = False, constraints: ChannelSpec | None = None
    ):
        """
        Adds specifications (placeholders) for barrier gate lines on quantum dot pairs.

        This method configures barrier gate line specifications (placeholders) for pairs of quantum dots.
        Barrier gates are used to control the tunnel coupling between quantum dots in a pair, enabling
        two-quantum-dot operations. The gates are configured as BARRIER_GATE line type with DC frequency.

        This is distinct from `add_voltage_gate_lines` which is a generic method. Use this method when
        you specifically need barrier gates for quantum dot pairs.

        No channels are allocated at this stage.

        Args:
            quantum_dot_pairs (QubitPairsType): The quantum dot pairs to configure the barrier gate lines for. Can be a single
                pair (tuple of two quantum dot IDs) or a list of pairs.
            triggered (bool, optional): Whether the line is triggered. Defaults to False.
            constraints (ChannelSpec, optional): Constraints on the channel, if any. Defaults to None.

        Returns:
            List[WiringSpec]: A list of wiring specifications (placeholders) for the barrier gate lines.
        """
        elements = self._make_qubit_pair_elements(quantum_dot_pairs)
        return self.add_wiring_spec(
            WiringFrequency.DC, WiringIOType.OUTPUT, WiringLineType.BARRIER_GATE, triggered, constraints, elements
        )
